@@ -13,7 +13,7 @@ extern crate test_utils;
 use solana_rbpf::{
     syscalls::BpfSyscallU64,
     user_error::UserError,
-    vm::{Config, DefaultInstructionMeter, Executable, SyscallObject, SyscallRegistry},
+    vm::{Config, Executable, SyscallObject, SyscallRegistry, TestInstructionMeter},
 };
 use std::{fs::File, io::Read};
 use test::Bencher;
@@ -32,7 +32,7 @@ fn bench_load_elf(bencher: &mut Bencher) {
     let mut elf = Vec::new();
     file.read_to_end(&mut elf).unwrap();
     bencher.iter(|| {
-        <dyn Executable<UserError, DefaultInstructionMeter>>::from_elf(
+        <dyn Executable<UserError, TestInstructionMeter>>::from_elf(
             &elf,
             None,
             Config::default(),
@@ -48,7 +48,7 @@ fn bench_load_elf_without_syscall(bencher: &mut Bencher) {
     let mut elf = Vec::new();
     file.read_to_end(&mut elf).unwrap();
     bencher.iter(|| {
-        let executable = <dyn Executable<UserError, DefaultInstructionMeter>>::from_elf(
+        let executable = <dyn Executable<UserError, TestInstructionMeter>>::from_elf(
             &elf,
             None,
             Config::default(),
@@ -65,7 +65,7 @@ fn bench_load_elf_with_syscall(bencher: &mut Bencher) {
     let mut elf = Vec::new();
     file.read_to_end(&mut elf).unwrap();
     bencher.iter(|| {
-        let executable = <dyn Executable<UserError, DefaultInstructionMeter>>::from_elf(
+        let executable = <dyn Executable<UserError, TestInstructionMeter>>::from_elf(
             &elf,
             None,
             Config::default(),
