@@ -62,9 +62,9 @@ pub enum ElfError {
     /// Incompatible ELF: wrong class
     #[error("Incompatible ELF: wrong class")]
     WrongClass,
-    /// Multiple text sections
-    #[error("Multiple text sections, consider removing llc option: -function-sections")]
-    MultipleTextSections,
+    /// Not one text section
+    #[error("Multiple or no text sections, consider removing llc option: -function-sections")]
+    NotOneTextSection,
     /// Read-write data not supported
     #[error("Found .bss section in ELF, read-write data not supported")]
     BssNotSupported,
@@ -88,7 +88,7 @@ pub enum ElfError {
     UnknownSymbol(usize),
     /// Offset or value is out of bounds
     #[error("Offset or value is out of bounds")]
-    OutOfBounds,
+    ValueOutOfBounds,
 }
 impl From<GoblinError> for ElfError {
     fn from(error: GoblinError) -> Self {
@@ -158,30 +158,41 @@ const BYTE_LENGTH_IMMEIDATE: usize = 4;
 enum BpfRelocationType {
     /// No relocation, placeholder
     R_Bpf_None = 0,
-    /// 64 bit relocation of a ldxdw instruction.
-    /// The ldxdw instruction occupies two instruction slots. The 64-bit address
-    /// to load from is split into the 32-bit imm field of each slot. The first
-    /// slot's pre-relocation imm field contains the virtual address (typically same
-    /// as the file offset) of the location to load. Relocation involves calculating
-    /// the post-load 64-bit physical address referenced by the imm field and writing
-    /// that physical address back into the imm fields of the ldxdw instruction.
+    /// R_BPF_64_64 relocation type is used for ld_imm64 instruction.
+    /// The actual to-be-relocated data (0 or section offset) is
+    /// stored at r_offset + 4 and the read/write data bitsize is 32
+    /// (4 bytes). The relocation can be resolved with the symbol
+    /// value plus implicit addend.
+    R_Bpf_64_64 = 1,
+    /// 64 bit relocation of a ldxdw instruction.  The ldxdw
+    /// instruction occupies two instruction slots. The 64-bit address
+    /// to load from is split into the 32-bit imm field of each
+    /// slot. The first slot's pre-relocation imm field contains the
+    /// virtual address (typically same as the file offset) of the
+    /// location to load. Relocation involves calculating the
+    /// post-load 64-bit physical address referenced by the imm field
+    /// and writing that physical address back into the imm fields of
+    /// the ldxdw instruction.
     R_Bpf_64_Relative = 8,
-    /// Relocation of a call instruction.
-    /// The existing imm field contains either an offset of the instruction to jump to
-    /// (think local function call) or a special value of "-1".  If -1 the symbol must
-    /// be looked up in the symbol table.  The relocation entry contains the symbol
-    /// number to call.  In order to support both local jumps and calling external
-    /// symbols a 32-bit hash is computed and stored in the the call instruction's
-    /// 32-bit imm field.  The hash is used later to look up the 64-bit address to
-    /// jump to.  In the case of a local jump the hash is calculated using the current
-    /// program counter and in the case of a symbol the hash is calculated using the
-    /// name of the symbol.
+    /// Relocation of a call instruction.  The existing imm field
+    /// contains either an offset of the instruction to jump to (think
+    /// local function call) or a special value of "-1".  If -1 the
+    /// symbol must be looked up in the symbol table.  The relocation
+    /// entry contains the symbol number to call.  In order to support
+    /// both local jumps and calling external symbols a 32-bit hash is
+    /// computed and stored in the the call instruction's 32-bit imm
+    /// field.  The hash is used later to look up the 64-bit address
+    /// to jump to.  In the case of a local jump the hash is
+    /// calculated using the current program counter and in the case
+    /// of a symbol the hash is calculated using the name of the
+    /// symbol.
     R_Bpf_64_32 = 10,
 }
 impl BpfRelocationType {
     fn from_x86_relocation_type(from: u32) -> Option<BpfRelocationType> {
         match from {
             R_X86_64_NONE => Some(BpfRelocationType::R_Bpf_None),
+            R_X86_64_64 => Some(BpfRelocationType::R_Bpf_64_64),
             R_X86_64_RELATIVE => Some(BpfRelocationType::R_Bpf_64_Relative),
             R_X86_64_32 => Some(BpfRelocationType::R_Bpf_64_32),
             _ => None,
@@ -203,10 +214,10 @@ pub struct EBpfElf<E: UserDefinedError, I: InstructionMeter> {
     config: Config,
     /// Loaded and executable elf
     elf_bytes: AlignedMemory,
+    /// Read-only section
+    ro_section: Vec<u8>,
     /// Text section info
     text_section_info: SectionInfo,
-    /// Read-only section info
-    ro_section_infos: Vec<SectionInfo>,
     /// Call resolution map (hash, pc, name)
     bpf_functions: BTreeMap<u32, (usize, String)>,
     /// Syscall symbol map (hash, name)
@@ -224,30 +235,17 @@ impl<E: UserDefinedError, I: InstructionMeter> Executable<E, I> for EBpfElf<E, I
     }
 
     /// Get the .text section virtual address and bytes
-    fn get_text_bytes(&self) -> Result<(u64, &[u8]), EbpfError<E>> {
-        Ok((
+    fn get_text_bytes(&self) -> (u64, &[u8]) {
+        let offset = (self.text_section_info.vaddr - ebpf::MM_PROGRAM_START) as usize;
+        (
             self.text_section_info.vaddr,
-            self.elf_bytes
-                .as_slice()
-                .get(self.text_section_info.offset_range.clone())
-                .ok_or(ElfError::OutOfBounds)?,
-        ))
+            &self.ro_section[offset..offset + self.text_section_info.offset_range.len()],
+        )
     }
 
-    /// Get a vector of virtual addresses for each read-only section
-    fn get_ro_sections(&self) -> Result<Vec<(u64, &[u8])>, EbpfError<E>> {
-        self.ro_section_infos
-            .iter()
-            .map(|section_info| {
-                Ok((
-                    section_info.vaddr,
-                    self.elf_bytes
-                        .as_slice()
-                        .get(section_info.offset_range.clone())
-                        .ok_or(ElfError::OutOfBounds)?,
-                ))
-            })
-            .collect::<Result<Vec<_>, EbpfError<E>>>()
+    /// Get the concatenated read-only sections (including the text section)
+    fn get_ro_section(&self) -> &[u8] {
+        self.ro_section.as_slice()
     }
 
     /// Get the entry point offset into the text section
@@ -288,19 +286,20 @@ impl<E: UserDefinedError, I: InstructionMeter> Executable<E, I> for EBpfElf<E, I
         let mut name = "Unknown";
         if let Ok(elf) = Elf::parse(self.elf_bytes.as_slice()) {
             for relocation in &elf.dynrels {
-                if let Some(BpfRelocationType::R_Bpf_64_32) =
-                    BpfRelocationType::from_x86_relocation_type(relocation.r_type)
-                {
-                    if relocation.r_offset as usize == file_offset {
-                        let sym = elf
-                            .dynsyms
-                            .get(relocation.r_sym)
-                            .ok_or(ElfError::UnknownSymbol(relocation.r_sym))?;
-                        name = elf
-                            .dynstrtab
-                            .get_at(sym.st_name)
-                            .ok_or(ElfError::UnknownSymbol(sym.st_name))?;
+                match BpfRelocationType::from_x86_relocation_type(relocation.r_type) {
+                    Some(BpfRelocationType::R_Bpf_64_32) | Some(BpfRelocationType::R_Bpf_64_64) => {
+                        if relocation.r_offset as usize == file_offset {
+                            let sym = elf
+                                .dynsyms
+                                .get(relocation.r_sym)
+                                .ok_or(ElfError::UnknownSymbol(relocation.r_sym))?;
+                            name = elf
+                                .dynstrtab
+                                .get_at(sym.st_name)
+                                .ok_or(ElfError::UnknownSymbol(sym.st_name))?;
+                        }
                     }
+                    _ => (),
                 }
             }
         }
@@ -339,6 +338,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
         Self {
             config,
             elf_bytes,
+            ro_section: text_bytes.to_vec(),
             text_section_info: SectionInfo {
                 name: ".text".to_string(),
                 vaddr: ebpf::MM_PROGRAM_START,
@@ -347,7 +347,6 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
                     end: text_bytes.len(),
                 },
             },
-            ro_section_infos: vec![],
             bpf_functions,
             syscall_symbols: BTreeMap::default(),
             syscall_registry,
@@ -378,7 +377,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
             offset_range: text_section.file_range().unwrap_or_default(),
         };
         if text_section_info.vaddr > ebpf::MM_STACK_START {
-            return Err(ElfError::OutOfBounds);
+            return Err(ElfError::ValueOutOfBounds);
         }
 
         // relocate symbols
@@ -402,36 +401,50 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
         bpf_functions.remove(&ebpf::hash_symbol_name(b"entrypoint"));
         register_bpf_function(&mut bpf_functions, entrypoint, "entrypoint")?;
 
-        // calculate the read-only section infos
-        let ro_section_infos = elf
+        // concatenate the read-only sections into one
+        let mut ro_length = text_section.sh_addr as usize + text_section_info.offset_range.len();
+        let ro_slices = elf
             .section_headers
             .iter()
-            .filter_map(|section_header| {
+            .filter(|section_header| {
                 if let Some(name) = elf.shdr_strtab.get_at(section_header.sh_name) {
-                    if name == ".rodata" || name == ".data.rel.ro" || name == ".eh_frame" {
-                        return Some(SectionInfo {
-                            name: name.to_string(),
-                            vaddr: section_header
-                                .sh_addr
-                                .saturating_add(ebpf::MM_PROGRAM_START),
-                            offset_range: section_header.file_range().unwrap_or_default(),
-                        });
-                    }
+                    return name == ".rodata" || name == ".data.rel.ro" || name == ".eh_frame";
                 }
-                None
+                false
             })
-            .collect::<Vec<_>>();
-        for ro_section_info in ro_section_infos.iter() {
-            if ro_section_info.vaddr > ebpf::MM_STACK_START {
-                return Err(ElfError::OutOfBounds);
-            }
+            .map(|section_header| {
+                let vaddr = section_header
+                    .sh_addr
+                    .saturating_add(ebpf::MM_PROGRAM_START);
+                if vaddr > ebpf::MM_STACK_START {
+                    return Err(ElfError::ValueOutOfBounds);
+                }
+                let slice = elf_bytes
+                    .as_slice()
+                    .get(section_header.file_range().unwrap_or_default())
+                    .ok_or(ElfError::ValueOutOfBounds)?;
+                ro_length = ro_length.max(section_header.sh_addr as usize + slice.len());
+                Ok((section_header.sh_addr as usize, slice))
+            })
+            .collect::<Result<Vec<_>, ElfError>>()?;
+        let mut ro_section = vec![0; ro_length];
+        ro_section[text_section.sh_addr as usize
+            ..text_section.sh_addr as usize + text_section_info.offset_range.len()]
+            .copy_from_slice(
+                elf_bytes
+                    .as_slice()
+                    .get(text_section_info.offset_range.clone())
+                    .ok_or(ElfError::ValueOutOfBounds)?,
+            );
+        for (offset, slice) in ro_slices.iter() {
+            ro_section[*offset..*offset + slice.len()].copy_from_slice(slice);
         }
 
         Ok(Self {
             config,
             elf_bytes,
+            ro_section,
             text_section_info,
-            ro_section_infos,
             bpf_functions,
             syscall_symbols,
             syscall_registry,
@@ -460,7 +473,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
                 insn.imm = hash as i64;
                 let checked_slice = elf_bytes
                     .get_mut(i * ebpf::INSN_SIZE..(i * ebpf::INSN_SIZE) + ebpf::INSN_SIZE)
-                    .ok_or(ElfError::OutOfBounds)?;
+                    .ok_or(ElfError::ValueOutOfBounds)?;
                 checked_slice.copy_from_slice(&insn.to_vec());
             }
         }
@@ -494,7 +507,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
             count
         });
         if 1 != num_text_sections {
-            return Err(ElfError::MultipleTextSections);
+            return Err(ElfError::NotOneTextSection);
         }
 
         for section_header in elf.section_headers.iter() {
@@ -510,8 +523,10 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
             let end = section_header
                 .sh_offset
                 .checked_add(section_header.sh_size)
-                .ok_or(ElfError::OutOfBounds)? as usize;
-            let _ = elf_bytes.get(start..end).ok_or(ElfError::OutOfBounds)?;
+                .ok_or(ElfError::ValueOutOfBounds)? as usize;
+            let _ = elf_bytes
+                .get(start..end)
+                .ok_or(ElfError::ValueOutOfBounds)?;
         }
         let text_section = Self::get_section(elf, ".text")?;
         if !text_section
@@ -553,9 +568,9 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
         // Fixup all program counter relative call instructions
         Self::fixup_relative_calls(
             bpf_functions,
-            &mut elf_bytes
+            elf_bytes
                 .get_mut(text_section.file_range().unwrap_or_default())
-                .ok_or(ElfError::OutOfBounds)?,
+                .ok_or(ElfError::ValueOutOfBounds)?,
         )?;
 
         let mut syscall_cache = BTreeMap::new();
@@ -568,6 +583,35 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
             // Offset of the immediate field
             let imm_offset = r_offset.saturating_add(BYTE_OFFSET_IMMEDIATE);
             match BpfRelocationType::from_x86_relocation_type(relocation.r_type) {
+                Some(BpfRelocationType::R_Bpf_64_64) => {
+                    // Read the instruction's immediate field which contains virtual
+                    // address to convert to physical
+                    let checked_slice = elf_bytes
+                        .get(imm_offset..imm_offset.saturating_add(BYTE_LENGTH_IMMEIDATE))
+                        .ok_or(ElfError::ValueOutOfBounds)?;
+                    let refd_va = LittleEndian::read_u32(checked_slice) as u64;
+                    // final "physical address" from the VM's perspetive is rooted at `MM_PROGRAM_START`
+                    let refd_pa = ebpf::MM_PROGRAM_START.saturating_add(refd_va);
+
+                    // The .text section has an unresolved load symbol instruction.
+                    let sym = elf
+                        .dynsyms
+                        .get(relocation.r_sym)
+                        .ok_or(ElfError::UnknownSymbol(relocation.r_sym))?;
+                    let addr = (sym.st_value + refd_pa) as u64;
+                    let checked_slice = elf_bytes
+                        .get_mut(imm_offset..imm_offset.saturating_add(BYTE_LENGTH_IMMEIDATE))
+                        .ok_or(ElfError::ValueOutOfBounds)?;
+                    LittleEndian::write_u32(checked_slice, (addr & 0xFFFFFFFF) as u32);
+                    let checked_slice = elf_bytes
+                        .get_mut(
+                            imm_offset.saturating_add(ebpf::INSN_SIZE)
+                                ..imm_offset
+                                    .saturating_add(ebpf::INSN_SIZE + BYTE_LENGTH_IMMEIDATE),
+                        )
+                        .ok_or(ElfError::ValueOutOfBounds)?;
+                    LittleEndian::write_u32(checked_slice, (addr >> 32) as u32);
+                }
                 Some(BpfRelocationType::R_Bpf_64_Relative) => {
                     // Raw relocation between sections.  The instruction being relocated contains
                     // the virtual address that it needs turned into a physical address.  Read it,
@@ -577,7 +621,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
                     // address to convert to physical
                     let checked_slice = elf_bytes
                         .get(imm_offset..imm_offset.saturating_add(BYTE_LENGTH_IMMEIDATE))
-                        .ok_or(ElfError::OutOfBounds)?;
+                        .ok_or(ElfError::ValueOutOfBounds)?;
                     let refd_va = LittleEndian::read_u32(checked_slice) as u64;
 
                     if refd_va == 0 {
@@ -596,24 +640,24 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
                         // Instruction lddw spans two instruction slots, split the
                         // physical address into a high and low and write into both slot's imm field
 
-                        let mut checked_slice = elf_bytes
+                        let checked_slice = elf_bytes
                             .get_mut(imm_offset..imm_offset.saturating_add(BYTE_LENGTH_IMMEIDATE))
-                            .ok_or(ElfError::OutOfBounds)?;
-                        LittleEndian::write_u32(&mut checked_slice, (refd_pa & 0xFFFFFFFF) as u32);
-                        let mut checked_slice = elf_bytes
+                            .ok_or(ElfError::ValueOutOfBounds)?;
+                        LittleEndian::write_u32(checked_slice, (refd_pa & 0xFFFFFFFF) as u32);
+                        let checked_slice = elf_bytes
                             .get_mut(
                                 imm_offset.saturating_add(ebpf::INSN_SIZE)
                                     ..imm_offset
                                         .saturating_add(ebpf::INSN_SIZE + BYTE_LENGTH_IMMEIDATE),
                             )
-                            .ok_or(ElfError::OutOfBounds)?;
-                        LittleEndian::write_u32(&mut checked_slice, (refd_pa >> 32) as u32);
+                            .ok_or(ElfError::ValueOutOfBounds)?;
+                        LittleEndian::write_u32(checked_slice, (refd_pa >> 32) as u32);
                     } else {
                         // 64 bit memory location, write entire 64 bit physical address directly
-                        let mut checked_slice = elf_bytes
+                        let checked_slice = elf_bytes
                             .get_mut(r_offset..r_offset.saturating_add(mem::size_of::<u64>()))
-                            .ok_or(ElfError::OutOfBounds)?;
-                        LittleEndian::write_u64(&mut checked_slice, refd_pa);
+                            .ok_or(ElfError::ValueOutOfBounds)?;
+                        LittleEndian::write_u64(checked_slice, refd_pa);
                     }
                 }
                 Some(BpfRelocationType::R_Bpf_64_32) => {
@@ -632,7 +676,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
                     let hash = if sym.is_function() && sym.st_value != 0 {
                         // bpf call
                         if !text_section.vm_range().contains(&(sym.st_value as usize)) {
-                            return Err(ElfError::OutOfBounds);
+                            return Err(ElfError::ValueOutOfBounds);
                         }
                         let target_pc =
                             (sym.st_value - text_section.sh_addr) as usize / ebpf::INSN_SIZE;
@@ -654,10 +698,10 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
                         }
                         hash
                     };
-                    let mut checked_slice = elf_bytes
+                    let checked_slice = elf_bytes
                         .get_mut(imm_offset..imm_offset.saturating_add(BYTE_LENGTH_IMMEIDATE))
-                        .ok_or(ElfError::OutOfBounds)?;
-                    LittleEndian::write_u32(&mut checked_slice, hash);
+                        .ok_or(ElfError::ValueOutOfBounds)?;
+                    LittleEndian::write_u32(checked_slice, hash);
                 }
                 _ => return Err(ElfError::UnknownRelocation(relocation.r_type)),
             }
@@ -678,7 +722,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EBpfElf<E, I> {
                 .vm_range()
                 .contains(&(symbol.st_value as usize))
             {
-                return Err(ElfError::OutOfBounds);
+                return Err(ElfError::ValueOutOfBounds);
             }
             let target_pc = (symbol.st_value - text_section.sh_addr) as usize / ebpf::INSN_SIZE;
             let name = elf
@@ -715,11 +759,11 @@ mod test {
         fuzz::fuzz,
         syscalls::{BpfSyscallString, BpfSyscallU64},
         user_error::UserError,
-        vm::{DefaultInstructionMeter, SyscallObject},
+        vm::{SyscallObject, TestInstructionMeter},
     };
     use rand::{distributions::Uniform, Rng};
     use std::{fs::File, io::Read};
-    type ElfExecutable = EBpfElf<UserError, DefaultInstructionMeter>;
+    type ElfExecutable = EBpfElf<UserError, TestInstructionMeter>;
 
     fn syscall_registry() -> SyscallRegistry {
         let mut syscall_registry = SyscallRegistry::default();
@@ -784,7 +828,7 @@ mod test {
             .expect("validation failed");
         let mut parsed_elf = Elf::parse(&elf_bytes).unwrap();
         let initial_e_entry = parsed_elf.header.e_entry;
-        let executable: &dyn Executable<UserError, DefaultInstructionMeter> = &elf;
+        let executable: &dyn Executable<UserError, TestInstructionMeter> = &elf;
         assert_eq!(
             0,
             executable
@@ -797,7 +841,7 @@ mod test {
         elf_bytes.pwrite(parsed_elf.header, 0).unwrap();
         let elf = ElfExecutable::load(Config::default(), &elf_bytes, syscall_registry())
             .expect("validation failed");
-        let executable: &dyn Executable<UserError, DefaultInstructionMeter> = &elf;
+        let executable: &dyn Executable<UserError, TestInstructionMeter> = &elf;
         assert_eq!(
             1,
             executable
@@ -834,7 +878,7 @@ mod test {
         elf_bytes.pwrite(parsed_elf.header, 0).unwrap();
         let elf = ElfExecutable::load(Config::default(), &elf_bytes, syscall_registry())
             .expect("validation failed");
-        let executable: &dyn Executable<UserError, DefaultInstructionMeter> = &elf;
+        let executable: &dyn Executable<UserError, TestInstructionMeter> = &elf;
         assert_eq!(
             0,
             executable
@@ -1048,4 +1092,14 @@ mod test {
             },
         );
     }
+
+    #[test]
+    fn test_relocs() {
+        let mut file = File::open("tests/elfs/reloc.so").expect("file open failed");
+        let mut elf_bytes = Vec::new();
+        file.read_to_end(&mut elf_bytes)
+            .expect("failed to read elf file");
+        ElfExecutable::load(Config::default(), &elf_bytes, syscall_registry())
+            .expect("validation failed");
+    }
 }
