@@ -20,13 +20,13 @@ use std::collections::HashMap;
 use std::fmt::Formatter;
 use std::fmt::Error as FormatterError;
 use std::ops::{Index, IndexMut};
-use rand::{rngs::ThreadRng, Rng};
+use rand::{rngs::SmallRng, Rng, SeedableRng};
 
 use crate::{
     vm::{Config, Executable, ProgramResult, InstructionMeter, Tracer, DynTraitFatPointer, SYSCALL_CONTEXT_OBJECTS_OFFSET, REPORT_UNRESOLVED_SYMBOL_INDEX},
     ebpf::{self, INSN_SIZE, FIRST_SCRATCH_REG, SCRATCH_REGS, STACK_REG, MM_STACK_START},
     error::{UserDefinedError, EbpfError},
-    memory_region::{AccessType, MemoryMapping},
+    memory_region::{AccessType, MemoryMapping, MemoryRegion},
     user_error::UserError,
     x86::*,
 };
@@ -146,7 +146,7 @@ impl<E: UserDefinedError, I: InstructionMeter> PartialEq for JitProgram<E, I> {
 
 impl<E: UserDefinedError, I: InstructionMeter> JitProgram<E, I> {
     pub fn new(executable: &dyn Executable<E, I>) -> Result<Self, EbpfError<E>> {
-        let program = executable.get_text_bytes()?.1;
+        let program = executable.get_text_bytes().1;
         let mut jit = JitCompiler::new::<E>(program, executable.get_config())?;
         jit.compile::<E, I>(executable)?;
         let main = unsafe { mem::transmute(jit.result.text_section.as_ptr()) };
@@ -158,9 +158,11 @@ impl<E: UserDefinedError, I: InstructionMeter> JitProgram<E, I> {
 }
 
 // Special values for target_pc in struct Jump
-const TARGET_PC_TRACE: usize = std::usize::MAX - 13;
-const TARGET_PC_TRANSLATE_PC: usize = std::usize::MAX - 12;
-const TARGET_PC_TRANSLATE_PC_LOOP: usize = std::usize::MAX - 11;
+const TARGET_PC_TRACE: usize = std::usize::MAX - 29;
+const TARGET_PC_TRANSLATE_PC: usize = std::usize::MAX - 28;
+const TARGET_PC_TRANSLATE_PC_LOOP: usize = std::usize::MAX - 27;
+const TARGET_PC_TRANSLATE_MEMORY_ADDRESS: usize = std::usize::MAX - 26;
+const TARGET_PC_MEMORY_ACCESS_VIOLATION: usize = std::usize::MAX - 18;
 const TARGET_PC_CALL_EXCEEDED_MAX_INSTRUCTIONS: usize = std::usize::MAX - 10;
 const TARGET_PC_CALL_DEPTH_EXCEEDED: usize = std::usize::MAX - 9;
 const TARGET_PC_CALL_OUTSIDE_TEXT_SEGMENT: usize = std::usize::MAX - 8;
@@ -168,7 +170,7 @@ const TARGET_PC_CALLX_UNSUPPORTED_INSTRUCTION: usize = std::usize::MAX - 7;
 const TARGET_PC_CALL_UNSUPPORTED_INSTRUCTION: usize = std::usize::MAX - 6;
 const TARGET_PC_DIV_BY_ZERO: usize = std::usize::MAX - 5;
 const TARGET_PC_EXCEPTION_AT: usize = std::usize::MAX - 4;
-const TARGET_PC_SYSCALL_EXCEPTION: usize = std::usize::MAX - 3;
+const TARGET_PC_RUST_EXCEPTION: usize = std::usize::MAX - 3;
 const TARGET_PC_EXIT: usize = std::usize::MAX - 2;
 const TARGET_PC_EPILOGUE: usize = std::usize::MAX - 1;
 
@@ -230,12 +232,12 @@ pub enum OperandSize {
 fn emit_sanitized_load_immediate<E: UserDefinedError>(jit: &mut JitCompiler, size: OperandSize, destination: u8, value: i64) -> Result<(), EbpfError<E>> {
     match size {
         OperandSize::S32 => {
-            let key: i32 = jit.rng.gen();
+            let key: i32 = jit.diversification_rng.gen();
             X86Instruction::load_immediate(size, destination, (value as i32).wrapping_sub(key) as i64).emit(jit)?;
             emit_alu(jit, size, 0x81, 0, destination, key as i64, None)
         },
         OperandSize::S64 if destination == R11 => {
-            let key: i64 = jit.rng.gen();
+            let key: i64 = jit.diversification_rng.gen();
             let lower_key = key as i32 as i64;
             let upper_key = (key >> 32) as i32 as i64;
             X86Instruction::load_immediate(size, destination, value.wrapping_sub(lower_key).rotate_right(32).wrapping_sub(upper_key)).emit(jit)?;
@@ -244,12 +246,12 @@ fn emit_sanitized_load_immediate<E: UserDefinedError>(jit: &mut JitCompiler, siz
             emit_alu(jit, size, 0x81, 0, destination, lower_key, None) // wrapping_add(lower_key)
         },
         OperandSize::S64 if value >= std::i32::MIN as i64 && value <= std::i32::MAX as i64 => {
-            let key = jit.rng.gen::<i32>() as i64;
+            let key = jit.diversification_rng.gen::<i32>() as i64;
             X86Instruction::load_immediate(size, destination, value.wrapping_sub(key)).emit(jit)?;
             emit_alu(jit, size, 0x81, 0, destination, key, None)
         },
         OperandSize::S64 => {
-            let key: i64 = jit.rng.gen();
+            let key: i64 = jit.diversification_rng.gen();
             X86Instruction::load_immediate(size, destination, value.wrapping_sub(key)).emit(jit)?;
             X86Instruction::load_immediate(size, R11, key).emit(jit)?;
             emit_alu(jit, size, 0x01, R11, destination, 0, None)
@@ -263,7 +265,7 @@ fn emit_sanitized_load_immediate<E: UserDefinedError>(jit: &mut JitCompiler, siz
 }
 
 fn emit_sanitized_load<E: UserDefinedError>(jit: &mut JitCompiler, size: OperandSize, source: u8, destination: u8, offset: i32) -> Result<(), EbpfError<E>> {
-    let key: i32 = jit.rng.gen();
+    let key: i32 = jit.diversification_rng.gen();
     X86Instruction::load_immediate(OperandSize::S64, destination, offset.wrapping_sub(key) as i64).emit(jit)?;
     X86Instruction::load(size, source, destination, X86IndirectAccess::OffsetIndexShift(key, destination, 0)).emit(jit)
 }
@@ -340,14 +342,39 @@ enum EnvironmentStackSlot {
     PrevInsnMeter = 8,
     /// Constant pointer to instruction_meter
     InsnMeterPtr = 9,
+    /// CPU cycles accumulated by the stop watch
+    StopwatchNumerator = 10,
+    /// Number of times the stop watch was used
+    StopwatchDenominator = 11,
     /// Bumper for size_of
-    SlotCount = 10,
+    SlotCount = 12,
 }
 
 fn slot_on_environment_stack(jit: &JitCompiler, slot: EnvironmentStackSlot) -> i32 {
     -8 * (slot as i32 + jit.environment_stack_key)
 }
 
+#[allow(dead_code)]
+#[inline]
+fn emit_stopwatch<E: UserDefinedError>(jit: &mut JitCompiler, begin: bool) -> Result<(), EbpfError<E>> {
+    jit.stopwatch_is_active = true;
+    X86Instruction::push(RDX).emit(jit)?;
+    X86Instruction::push(RAX).emit(jit)?;
+    X86Instruction::fence(FenceType::Load).emit(jit)?; // lfence
+    X86Instruction::cycle_count().emit(jit)?; // rdtsc
+    X86Instruction::fence(FenceType::Load).emit(jit)?; // lfence
+    emit_alu(jit, OperandSize::S64, 0xc1, 4, RDX, 32, None)?; // RDX <<= 32;
+    emit_alu(jit, OperandSize::S64, 0x09, RDX, RAX, 0, None)?; // RAX |= RDX;
+    if begin {
+        emit_alu(jit, OperandSize::S64, 0x29, RAX, RBP, 0, Some(X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::StopwatchNumerator))))?; // *numerator -= RAX;
+    } else {
+        emit_alu(jit, OperandSize::S64, 0x01, RAX, RBP, 0, Some(X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::StopwatchNumerator))))?; // *numerator += RAX;
+        emit_alu(jit, OperandSize::S64, 0x81, 0, RBP, 1, Some(X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::StopwatchDenominator))))?; // *denominator += 1;
+    }
+    X86Instruction::pop(RAX).emit(jit)?;
+    X86Instruction::pop(RDX).emit(jit)
+}
+
 /* Explaination of the Instruction Meter
 
     The instruction meter serves two purposes: First, measure how many BPF instructions are
@@ -397,19 +424,28 @@ fn slot_on_environment_stack(jit: &JitCompiler, slot: EnvironmentStackSlot) -> i
     and undo again can be anything, so we just set it to zero.
 */
 
+#[inline]
+fn emit_validate_instruction_count<E: UserDefinedError>(jit: &mut JitCompiler, exclusive: bool, pc: Option<usize>) -> Result<(), EbpfError<E>> {
+    if let Some(pc) = pc {
+        jit.last_instruction_meter_validation_pc = pc;
+        X86Instruction::cmp_immediate(OperandSize::S64, ARGUMENT_REGISTERS[0], pc as i64 + 1, None).emit(jit)?;
+    } else {
+        X86Instruction::cmp(OperandSize::S64, R11, ARGUMENT_REGISTERS[0], None).emit(jit)?;
+    }
+    emit_jcc(jit, if exclusive { 0x82 } else { 0x86 }, TARGET_PC_CALL_EXCEEDED_MAX_INSTRUCTIONS)
+}
+
 #[inline]
 fn emit_profile_instruction_count<E: UserDefinedError>(jit: &mut JitCompiler, target_pc: Option<usize>) -> Result<(), EbpfError<E>> {
-    if jit.config.enable_instruction_meter {
-        match target_pc {
-            Some(target_pc) => {
-                emit_alu(jit, OperandSize::S64, 0x81, 0, ARGUMENT_REGISTERS[0], target_pc as i64 - jit.pc as i64 - 1, None)?; // instruction_meter += target_pc - (jit.pc + 1);
-            },
-            None => { // If no constant target_pc is given, it is expected to be on the stack instead
-                X86Instruction::pop(R11).emit(jit)?;
-                emit_alu(jit, OperandSize::S64, 0x81, 5, ARGUMENT_REGISTERS[0], jit.pc as i64 + 1, None)?; // instruction_meter -= jit.pc + 1;
-                emit_alu(jit, OperandSize::S64, 0x01, R11, ARGUMENT_REGISTERS[0], jit.pc as i64, None)?; // instruction_meter += target_pc;
-            },
-        }
+    match target_pc {
+        Some(target_pc) => {
+            emit_alu(jit, OperandSize::S64, 0x81, 0, ARGUMENT_REGISTERS[0], target_pc as i64 - jit.pc as i64 - 1, None)?; // instruction_meter += target_pc - (jit.pc + 1);
+        },
+        None => { // If no constant target_pc is given, it is expected to be on the stack instead
+            X86Instruction::pop(R11).emit(jit)?;
+            emit_alu(jit, OperandSize::S64, 0x81, 5, ARGUMENT_REGISTERS[0], jit.pc as i64 + 1, None)?; // instruction_meter -= jit.pc + 1;
+            emit_alu(jit, OperandSize::S64, 0x01, R11, ARGUMENT_REGISTERS[0], jit.pc as i64, None)?; // instruction_meter += target_pc;
+        },
     }
     Ok(())
 }
@@ -417,8 +453,7 @@ fn emit_profile_instruction_count<E: UserDefinedError>(jit: &mut JitCompiler, ta
 #[inline]
 fn emit_validate_and_profile_instruction_count<E: UserDefinedError>(jit: &mut JitCompiler, exclusive: bool, target_pc: Option<usize>) -> Result<(), EbpfError<E>> {
     if jit.config.enable_instruction_meter {
-        X86Instruction::cmp_immediate(OperandSize::S64, ARGUMENT_REGISTERS[0], jit.pc as i64 + 1, None).emit(jit)?;
-        emit_jcc(jit, if exclusive { 0x82 } else { 0x86 }, TARGET_PC_CALL_EXCEEDED_MAX_INSTRUCTIONS)?;
+        emit_validate_instruction_count(jit, exclusive, Some(jit.pc))?;
         emit_profile_instruction_count(jit, target_pc)?;
     }
     Ok(())
@@ -433,11 +468,17 @@ fn emit_undo_profile_instruction_count<E: UserDefinedError>(jit: &mut JitCompile
 }
 
 #[inline]
-fn emit_profile_instruction_count_of_exception<E: UserDefinedError>(jit: &mut JitCompiler) -> Result<(), EbpfError<E>> {
+fn emit_profile_instruction_count_of_exception<E: UserDefinedError>(jit: &mut JitCompiler, store_pc_in_exception: bool) -> Result<(), EbpfError<E>> {
     emit_alu(jit, OperandSize::S64, 0x81, 0, R11, 1, None)?;
     if jit.config.enable_instruction_meter {
         emit_alu(jit, OperandSize::S64, 0x29, R11, ARGUMENT_REGISTERS[0], 0, None)?; // instruction_meter -= pc + 1;
     }
+    if store_pc_in_exception {
+        X86Instruction::load(OperandSize::S64, RBP, R10, X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::OptRetValPtr))).emit(jit)?;
+        X86Instruction::store_immediate(OperandSize::S64, R10, X86IndirectAccess::Offset(0), 1).emit(jit)?; // is_err = true;
+        emit_alu(jit, OperandSize::S64, 0x81, 0, R11, ebpf::ELF_INSN_DUMP_OFFSET as i64 - 1, None)?;
+        X86Instruction::store(OperandSize::S64, R11, R10, X86IndirectAccess::Offset(16)).emit(jit)?; // pc = jit.pc + ebpf::ELF_INSN_DUMP_OFFSET;
+    }
     Ok(())
 }
 
@@ -520,7 +561,7 @@ fn emit_bpf_call<E: UserDefinedError>(jit: &mut JitCompiler, dst: Value, number_
                 emit_alu(jit, OperandSize::S64, 0xc1, 5, REGISTER_MAP[STACK_REG], shift_amount as i64, None)?;
                 X86Instruction::push(REGISTER_MAP[STACK_REG]).emit(jit)?;
             }
-            // Load host target_address from JitProgramArgument.instruction_addresses
+            // Load host target_address from jit_program_argument.instruction_addresses
             debug_assert_eq!(INSN_SIZE, 8); // Because the instruction size is also the slot size we do not need to shift the offset
             X86Instruction::mov(OperandSize::S64, REGISTER_MAP[0], REGISTER_MAP[STACK_REG]).emit(jit)?;
             X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[STACK_REG], jit.result.pc_section.as_ptr() as i64).emit(jit)?;
@@ -534,13 +575,13 @@ fn emit_bpf_call<E: UserDefinedError>(jit: &mut JitCompiler, dst: Value, number_
         }
     }
 
+    let stack_frame_size = jit.config.stack_frame_size as i64 * if jit.config.enable_stack_frame_gaps { 2 } else { 1 };
     X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[STACK_REG], X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::BpfStackPtr))).emit(jit)?;
-    emit_alu(jit, OperandSize::S64, 0x81, 4, REGISTER_MAP[STACK_REG], !(jit.config.stack_frame_size as i64 * 2 - 1), None)?; // stack_ptr &= !(jit.config.stack_frame_size * 2 - 1);
-    emit_alu(jit, OperandSize::S64, 0x81, 0, REGISTER_MAP[STACK_REG], jit.config.stack_frame_size as i64 * 3, None)?; // stack_ptr += jit.config.stack_frame_size * 3;
+    emit_alu(jit, OperandSize::S64, 0x81, 0, REGISTER_MAP[STACK_REG], stack_frame_size, None)?; // stack_ptr += stack_frame_increment;
     X86Instruction::store(OperandSize::S64, REGISTER_MAP[STACK_REG], RBP, X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::BpfStackPtr))).emit(jit)?;
 
-    // if(stack_ptr >= MM_STACK_START + jit.config.max_call_depth * jit.config.stack_frame_size * 2) throw EbpfError::CallDepthExeeded;
-    X86Instruction::load_immediate(OperandSize::S64, R11, MM_STACK_START as i64 + (jit.config.max_call_depth * jit.config.stack_frame_size * 2) as i64).emit(jit)?;
+    // if(stack_ptr >= MM_STACK_START + jit.config.stack_frame_size + jit.config.max_call_depth * stack_frame_size) throw EbpfError::CallDepthExeeded;
+    X86Instruction::load_immediate(OperandSize::S64, R11, MM_STACK_START as i64 + jit.config.stack_frame_size as i64 + (jit.config.max_call_depth as i64 * stack_frame_size)).emit(jit)?;
     X86Instruction::cmp(OperandSize::S64, R11, REGISTER_MAP[STACK_REG], None).emit(jit)?;
     // Store PC in case the bounds check fails
     X86Instruction::load_immediate(OperandSize::S64, R11, jit.pc as i64).emit(jit)?;
@@ -693,21 +734,29 @@ fn emit_rust_call<E: UserDefinedError>(jit: &mut JitCompiler, function: *const u
 
 #[inline]
 fn emit_address_translation<E: UserDefinedError>(jit: &mut JitCompiler, host_addr: u8, vm_addr: Value, len: u64, access_type: AccessType) -> Result<(), EbpfError<E>> {
-    emit_rust_call(jit, MemoryMapping::map::<UserError> as *const u8, &[
-        Argument { index: 3, value: vm_addr }, // Specify first as the src register could be overwritten by other arguments
-        Argument { index: 4, value: Value::Constant64(len as i64, false) },
-        Argument { index: 2, value: Value::Constant64(access_type as i64, false) },
-        Argument { index: 1, value: Value::RegisterPlusConstant32(R10, jit.program_argument_key, false) }, // JitProgramArgument::memory_mapping
-        Argument { index: 0, value: Value::RegisterIndirect(RBP, slot_on_environment_stack(jit, EnvironmentStackSlot::OptRetValPtr), false) },
-    ], None, true)?;
-
-    // Throw error if the result indicates one
-    X86Instruction::load_immediate(OperandSize::S64, R11, jit.pc as i64).emit(jit)?;
-    emit_jcc(jit, 0x85, TARGET_PC_EXCEPTION_AT)?;
-
-    // Store Ok value in result register
-    X86Instruction::load(OperandSize::S64, RBP, R11, X86IndirectAccess::Offset(slot_on_environment_stack(jit, EnvironmentStackSlot::OptRetValPtr))).emit(jit)?;
-    X86Instruction::load(OperandSize::S64, R11, host_addr, X86IndirectAccess::Offset(8)).emit(jit)
+    match vm_addr {
+        Value::RegisterPlusConstant64(reg, constant, user_provided) => {
+            if user_provided && jit.config.sanitize_user_provided_values {
+                emit_sanitized_load_immediate(jit, OperandSize::S64, R11, constant)?;
+            } else {
+                X86Instruction::load_immediate(OperandSize::S64, R11, constant).emit(jit)?;
+            }
+            emit_alu(jit, OperandSize::S64, 0x01, reg, R11, 0, None)?;
+        },
+        Value::Constant64(constant, user_provided) => {
+            if user_provided && jit.config.sanitize_user_provided_values {
+                emit_sanitized_load_immediate(jit, OperandSize::S64, R11, constant)?;
+            } else {
+                X86Instruction::load_immediate(OperandSize::S64, R11, constant).emit(jit)?;
+            }
+        },
+        _ => {
+            #[cfg(debug_assertions)]
+            unreachable!();
+        },
+    }
+    emit_call(jit, TARGET_PC_TRANSLATE_MEMORY_ADDRESS + len.trailing_zeros() as usize + 4 * (access_type as usize))?;
+    X86Instruction::mov(OperandSize::S64, R11, host_addr).emit(jit)
 }
 
 fn emit_shift<E: UserDefinedError>(jit: &mut JitCompiler, size: OperandSize, opcode_extension: u8, source: u8, destination: u8, immediate: Option<i64>) -> Result<(), EbpfError<E>> {
@@ -757,11 +806,7 @@ fn emit_muldivmod<E: UserDefinedError>(jit: &mut JitCompiler, opc: u8, src: u8,
     if (div || modrm) && imm.is_none() {
         // Save pc
         X86Instruction::load_immediate(OperandSize::S64, R11, jit.pc as i64).emit(jit)?;
-
-        // test src,src
-        emit_alu(jit, size, 0x85, src, src, 0, None)?;
-
-        // Jump if src is zero
+        X86Instruction::test(size, src, src, None).emit(jit)?; // src == 0
         emit_jcc(jit, 0x84, TARGET_PC_DIV_BY_ZERO)?;
     }
 
@@ -840,10 +885,12 @@ pub struct JitCompiler {
     text_section_jumps: Vec<Jump>,
     offset_in_text_section: usize,
     pc: usize,
+    last_instruction_meter_validation_pc: usize,
     program_vm_addr: u64,
     handler_anchors: HashMap<usize, usize>,
     config: Config,
-    rng: ThreadRng,
+    diversification_rng: SmallRng,
+    stopwatch_is_active: bool,
     environment_stack_key: i32,
     program_argument_key: i32,
 }
@@ -904,12 +951,12 @@ impl JitCompiler {
             };
         }
 
-        let mut code_length_estimate = pc * 256 + 512;
+        let mut code_length_estimate = pc * 256 + 4096;
         code_length_estimate += (code_length_estimate as f64 * _config.noop_instruction_ratio) as usize;
-        let mut rng = rand::thread_rng();
+        let mut diversification_rng = SmallRng::from_rng(rand::thread_rng()).unwrap();
         let (environment_stack_key, program_argument_key) =
             if _config.encrypt_environment_registers {
-                (rng.gen::<i32>() / 8, rng.gen())
+                (diversification_rng.gen::<i32>() / 8, diversification_rng.gen())
             } else { (0, 0) };
 
         Ok(Self {
@@ -918,10 +965,12 @@ impl JitCompiler {
             text_section_jumps: vec![],
             offset_in_text_section: 0,
             pc: 0,
+            last_instruction_meter_validation_pc: 0,
             program_vm_addr: 0,
             handler_anchors: HashMap::new(),
             config: *_config,
-            rng,
+            diversification_rng,
+            stopwatch_is_active: false,
             environment_stack_key,
             program_argument_key,
         })
@@ -929,14 +978,16 @@ impl JitCompiler {
 
     fn compile<E: UserDefinedError, I: InstructionMeter>(&mut self,
             executable: &dyn Executable<E, I>) -> Result<(), EbpfError<E>> {
-        let (program_vm_addr, program) = executable.get_text_bytes()?;
+        let (program_vm_addr, program) = executable.get_text_bytes();
         self.program_vm_addr = program_vm_addr;
 
         self.generate_prologue::<E, I>()?;
 
         // Jump to entry point
         let entry = executable.get_entrypoint_instruction_offset().unwrap_or(0);
-        emit_profile_instruction_count(self, Some(entry + 1))?;
+        if self.config.enable_instruction_meter {
+            emit_profile_instruction_count(self, Some(entry + 1))?;
+        }
         X86Instruction::load_immediate(OperandSize::S64, R11, entry as i64).emit(self)?;
         emit_jmp(self, entry)?;
 
@@ -949,6 +1000,11 @@ impl JitCompiler {
 
             self.result.pc_section[self.pc] = self.offset_in_text_section as u64;
 
+            // Regular instruction meter checkpoints to prevent long linear runs from exceeding their budget
+            if self.last_instruction_meter_validation_pc + self.config.instruction_meter_checkpoint_distance <= self.pc {
+                emit_validate_instruction_count(self, true, Some(self.pc))?;
+            }
+
             if self.config.enable_instruction_tracing {
                 X86Instruction::load_immediate(OperandSize::S64, R11, self.pc as i64).emit(self)?;
                 emit_call(self, TARGET_PC_TRACE)?;
@@ -1209,22 +1265,14 @@ impl JitCompiler {
                         X86Instruction::load(OperandSize::S64, R10, RAX, X86IndirectAccess::Offset((SYSCALL_CONTEXT_OBJECTS_OFFSET + syscall.context_object_slot) as i32 * 8 + self.program_argument_key)).emit(self)?;
                         emit_rust_call(self, syscall.function as *const u8, &[
                             Argument { index: 7, value: Value::RegisterIndirect(RBP, slot_on_environment_stack(self, EnvironmentStackSlot::OptRetValPtr), false) },
-                            Argument { index: 6, value: Value::RegisterPlusConstant32(R10, self.program_argument_key, false) }, // JitProgramArgument::memory_mapping
+                            Argument { index: 6, value: Value::RegisterPlusConstant32(R10, self.program_argument_key, false) }, // jit_program_argument.memory_mapping
                             Argument { index: 5, value: Value::Register(ARGUMENT_REGISTERS[5]) },
                             Argument { index: 4, value: Value::Register(ARGUMENT_REGISTERS[4]) },
                             Argument { index: 3, value: Value::Register(ARGUMENT_REGISTERS[3]) },
                             Argument { index: 2, value: Value::Register(ARGUMENT_REGISTERS[2]) },
                             Argument { index: 1, value: Value::Register(ARGUMENT_REGISTERS[1]) },
                             Argument { index: 0, value: Value::Register(RAX) }, // "&mut self" in the "call" method of the SyscallObject
-                        ], None, true)?;
-
-                        // Throw error if the result indicates one
-                        X86Instruction::load_immediate(OperandSize::S64, R11, self.pc as i64).emit(self)?;
-                        emit_jcc(self, 0x85, TARGET_PC_SYSCALL_EXCEPTION)?;
-
-                        // Store Ok value in result register
-                        X86Instruction::load(OperandSize::S64, RBP, R11, X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::OptRetValPtr))).emit(self)?;
-                        X86Instruction::load(OperandSize::S64, R11, REGISTER_MAP[0], X86IndirectAccess::Offset(8)).emit(self)?;
+                        ], None, false)?;
 
                         if self.config.enable_instruction_meter {
                             X86Instruction::load(OperandSize::S64, RBP, R11, X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::InsnMeterPtr))).emit(self)?;
@@ -1234,6 +1282,15 @@ impl JitCompiler {
                             X86Instruction::store(OperandSize::S64, ARGUMENT_REGISTERS[0], RBP, X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::PrevInsnMeter))).emit(self)?;
                             emit_undo_profile_instruction_count(self, 0)?;
                         }
+
+                        // Store Ok value in result register
+                        X86Instruction::load(OperandSize::S64, RBP, R11, X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::OptRetValPtr))).emit(self)?;
+                        X86Instruction::load(OperandSize::S64, R11, REGISTER_MAP[0], X86IndirectAccess::Offset(8)).emit(self)?;
+
+                        // Throw error if the result indicates one
+                        X86Instruction::cmp_immediate(OperandSize::S64, R11, 0, Some(X86IndirectAccess::Offset(0))).emit(self)?;
+                        X86Instruction::load_immediate(OperandSize::S64, R11, self.pc as i64).emit(self)?;
+                        emit_jcc(self, 0x85, TARGET_PC_RUST_EXCEPTION)?;
                     } else if let Some(target_pc) = executable.lookup_bpf_function(insn.imm as u32) {
                         emit_bpf_call(self, Value::Constant64(target_pc as i64, false), self.result.pc_section.len() - 1)?;
                     } else {
@@ -1246,7 +1303,8 @@ impl JitCompiler {
                             Argument { index: 0, value: Value::RegisterIndirect(RBP, slot_on_environment_stack(self, EnvironmentStackSlot::OptRetValPtr), false) },
                         ], None, true)?;
                         X86Instruction::load_immediate(OperandSize::S64, R11, self.pc as i64).emit(self)?;
-                        emit_jmp(self, TARGET_PC_SYSCALL_EXCEPTION)?;
+                        emit_validate_instruction_count(self, false, None)?;
+                        emit_jmp(self, TARGET_PC_RUST_EXCEPTION)?;
                     }
                 },
                 ebpf::CALL_REG  => {
@@ -1255,16 +1313,14 @@ impl JitCompiler {
                 ebpf::EXIT      => {
                     emit_validate_and_profile_instruction_count(self, true, Some(0))?;
 
+                    let stack_frame_size = self.config.stack_frame_size as i64 * if self.config.enable_stack_frame_gaps { 2 } else { 1 };
                     X86Instruction::load(OperandSize::S64, RBP, REGISTER_MAP[STACK_REG], X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfStackPtr))).emit(self)?;
-                    emit_alu(self, OperandSize::S64, 0x81, 4, REGISTER_MAP[STACK_REG], !(self.config.stack_frame_size as i64 * 2 - 1), None)?; // stack_ptr &= !(jit.config.stack_frame_size * 2 - 1);
-                    emit_alu(self, OperandSize::S64, 0x81, 5, REGISTER_MAP[STACK_REG], self.config.stack_frame_size as i64 * 2, None)?; // stack_ptr -= jit.config.stack_frame_size * 2;
+                    emit_alu(self, OperandSize::S64, 0x81, 5, REGISTER_MAP[STACK_REG], stack_frame_size, None)?; // stack_ptr -= stack_frame_size;
                     X86Instruction::store(OperandSize::S64, REGISTER_MAP[STACK_REG], RBP, X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::BpfStackPtr))).emit(self)?;
 
-                    // if(stack_ptr < MM_STACK_START) goto exit;
-                    X86Instruction::mov(OperandSize::S64, REGISTER_MAP[0], R11).emit(self)?;
-                    X86Instruction::load_immediate(OperandSize::S64, REGISTER_MAP[0], MM_STACK_START as i64).emit(self)?;
-                    X86Instruction::cmp(OperandSize::S64, REGISTER_MAP[0], REGISTER_MAP[STACK_REG], None).emit(self)?;
-                    X86Instruction::mov(OperandSize::S64, R11, REGISTER_MAP[0]).emit(self)?;
+                    // if(stack_ptr < MM_STACK_START + self.config.stack_frame_size) goto exit;
+                    X86Instruction::load_immediate(OperandSize::S64, R11, MM_STACK_START as i64 + self.config.stack_frame_size as i64).emit(self)?;
+                    X86Instruction::cmp(OperandSize::S64, R11, REGISTER_MAP[STACK_REG], None).emit(self)?;
                     emit_jcc(self, 0x82, TARGET_PC_EXIT)?;
 
                     // else return;
@@ -1331,7 +1387,87 @@ impl JitCompiler {
         emit_alu(self, OperandSize::S64, 0x29, REGISTER_MAP[0], R11, 0, None)?; // R11 -= REGISTER_MAP[0];
         emit_alu(self, OperandSize::S64, 0xc1, 5, R11, 3, None)?; // R11 >>= 3;
         X86Instruction::pop(REGISTER_MAP[0]).emit(self)?; // Restore REGISTER_MAP[0]
-        X86Instruction::return_near().emit(self)
+        X86Instruction::return_near().emit(self)?;
+
+        // Translates a vm memory address to a host memory address
+        for (access_type, len) in &[
+            (AccessType::Load, 1i32),
+            (AccessType::Load, 2i32),
+            (AccessType::Load, 4i32),
+            (AccessType::Load, 8i32),
+            (AccessType::Store, 1i32),
+            (AccessType::Store, 2i32),
+            (AccessType::Store, 4i32),
+            (AccessType::Store, 8i32),
+        ] {
+            let target_offset = len.trailing_zeros() as usize + 4 * (*access_type as usize);
+
+            set_anchor(self, TARGET_PC_TRANSLATE_MEMORY_ADDRESS + target_offset);
+            X86Instruction::push(R11).emit(self)?;
+            X86Instruction::push(RAX).emit(self)?;
+            X86Instruction::push(RCX).emit(self)?;
+            let stack_offset = if self.config.enable_stack_frame_gaps {
+                X86Instruction::push(RDX).emit(self)?;
+                24
+            } else {
+                16
+            };
+            X86Instruction::mov(OperandSize::S64, R11, RAX).emit(self)?; // RAX = vm_addr;
+            emit_alu(self, OperandSize::S64, 0xc1, 5, RAX, ebpf::VIRTUAL_ADDRESS_BITS as i64, None)?; // RAX >>= ebpf::VIRTUAL_ADDRESS_BITS;
+            X86Instruction::cmp(OperandSize::S64, RAX, R10, Some(X86IndirectAccess::Offset(self.program_argument_key + 8))).emit(self)?; // region_index >= jit_program_argument.memory_mapping.regions.len()
+            emit_jcc(self, 0x86, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset)?;
+            debug_assert_eq!(1 << 5, std::mem::size_of::<MemoryRegion>());
+            emit_alu(self, OperandSize::S64, 0xc1, 4, RAX, 5, None)?; // RAX *= std::mem::size_of::<MemoryRegion>();
+            emit_alu(self, OperandSize::S64, 0x03, RAX, R10, 0, Some(X86IndirectAccess::Offset(self.program_argument_key)))?; // region = &jit_program_argument.memory_mapping.regions[region_index];
+            if *access_type == AccessType::Store {
+                X86Instruction::cmp_immediate(OperandSize::S8, RAX, 0, Some(X86IndirectAccess::Offset(25))).emit(self)?; // region.is_writable == 0
+                emit_jcc(self, 0x84, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset)?;
+            }
+            X86Instruction::load_immediate(OperandSize::S64, RCX, (1i64 << ebpf::VIRTUAL_ADDRESS_BITS) - 1).emit(self)?; // RCX = (1 << ebpf::VIRTUAL_ADDRESS_BITS) - 1;
+            emit_alu(self, OperandSize::S64, 0x21, RCX, R11, 0, None)?; // R11 &= (1 << ebpf::VIRTUAL_ADDRESS_BITS) - 1;
+            if self.config.enable_stack_frame_gaps {
+                X86Instruction::load(OperandSize::S8, RAX, RCX, X86IndirectAccess::Offset(24)).emit(self)?; // RCX = region.vm_gap_shift;
+                X86Instruction::mov(OperandSize::S64, R11, RDX).emit(self)?; // RDX = R11;
+                emit_alu(self, OperandSize::S64, 0xd3, 5, RDX, 0, None)?; // RDX = R11 >> region.vm_gap_shift;
+                X86Instruction::test_immediate(OperandSize::S64, RDX, 1, None).emit(self)?; // (RDX & 1) != 0
+                emit_jcc(self, 0x85, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset)?;
+                X86Instruction::load_immediate(OperandSize::S64, RDX, -1).emit(self)?; // RDX = -1;
+                emit_alu(self, OperandSize::S64, 0xd3, 4, RDX, 0, None)?; // gap_mask = -1 << region.vm_gap_shift;
+                X86Instruction::mov(OperandSize::S64, RDX, RCX).emit(self)?; // RCX = RDX;
+                emit_alu(self, OperandSize::S64, 0xf7, 2, RCX, 0, None)?; // inverse_gap_mask = !gap_mask;
+                emit_alu(self, OperandSize::S64, 0x21, R11, RCX, 0, None)?; // below_gap = R11 & inverse_gap_mask;
+                emit_alu(self, OperandSize::S64, 0x21, RDX, R11, 0, None)?; // above_gap = R11 & gap_mask;
+                emit_alu(self, OperandSize::S64, 0xc1, 5, R11, 1, None)?; // above_gap >>= 1;
+                emit_alu(self, OperandSize::S64, 0x09, RCX, R11, 0, None)?; // gapped_offset = above_gap | below_gap;
+            }
+            X86Instruction::lea(OperandSize::S64, R11, RCX, Some(X86IndirectAccess::Offset(*len))).emit(self)?; // RCX = R11 + len;
+            X86Instruction::cmp(OperandSize::S8, RCX, RAX, Some(X86IndirectAccess::Offset(16))).emit(self)?; // region.len < R11 + len
+            emit_jcc(self, 0x82, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset)?;
+            emit_alu(self, OperandSize::S64, 0x03, R11, RAX, 0, Some(X86IndirectAccess::Offset(0)))?; // R11 += region.host_addr;
+            if self.config.enable_stack_frame_gaps {
+                X86Instruction::pop(RDX).emit(self)?;
+            }
+            X86Instruction::pop(RCX).emit(self)?;
+            X86Instruction::pop(RAX).emit(self)?;
+            emit_alu(self, OperandSize::S64, 0x81, 0, RSP, 8, None)?;
+            X86Instruction::return_near().emit(self)?;
+
+            set_anchor(self, TARGET_PC_MEMORY_ACCESS_VIOLATION + target_offset);
+            emit_alu(self, OperandSize::S64, 0x31, R11, R11, 0, None)?; // R11 = 0;
+            X86Instruction::load(OperandSize::S64, RSP, R11, X86IndirectAccess::OffsetIndexShift(stack_offset, R11, 0)).emit(self)?;
+            emit_rust_call(self, MemoryMapping::generate_access_violation::<UserError> as *const u8, &[
+                Argument { index: 3, value: Value::Register(R11) }, // Specify first as the src register could be overwritten by other arguments
+                Argument { index: 4, value: Value::Constant64(*len as i64, false) },
+                Argument { index: 2, value: Value::Constant64(*access_type as i64, false) },
+                Argument { index: 1, value: Value::RegisterPlusConstant32(R10, self.program_argument_key, false) }, // jit_program_argument.memory_mapping
+                Argument { index: 0, value: Value::RegisterIndirect(RBP, slot_on_environment_stack(self, EnvironmentStackSlot::OptRetValPtr), false) }, // Pointer to optional typed return value
+            ], None, true)?;
+            emit_alu(self, OperandSize::S64, 0x81, 0, RSP, stack_offset as i64 + 8, None)?; // Drop R11, RAX, RCX, RDX from stack
+            X86Instruction::pop(R11).emit(self)?; // Put callers PC in R11
+            emit_call(self, TARGET_PC_TRANSLATE_PC)?;
+            emit_jmp(self, TARGET_PC_EXCEPTION_AT)?;
+        }
+        Ok(())
     }
 
     fn generate_exception_handlers<E: UserDefinedError>(&mut self) -> Result<(), EbpfError<E>> {
@@ -1339,7 +1475,8 @@ impl JitCompiler {
         set_anchor(self, TARGET_PC_CALL_EXCEEDED_MAX_INSTRUCTIONS);
         X86Instruction::mov(OperandSize::S64, ARGUMENT_REGISTERS[0], R11).emit(self)?;
         emit_set_exception_kind::<E>(self, EbpfError::ExceededMaxInstructions(0, 0))?;
-        emit_jmp(self, TARGET_PC_EXCEPTION_AT)?;
+        emit_profile_instruction_count_of_exception(self, true)?;
+        emit_jmp(self, TARGET_PC_EPILOGUE)?;
 
         // Handler for EbpfError::CallDepthExceeded
         set_anchor(self, TARGET_PC_CALL_DEPTH_EXCEEDED);
@@ -1376,16 +1513,16 @@ impl JitCompiler {
 
         // Handler for exceptions which report their pc
         set_anchor(self, TARGET_PC_EXCEPTION_AT);
-        emit_profile_instruction_count_of_exception(self)?;
-        X86Instruction::load(OperandSize::S64, RBP, R10, X86IndirectAccess::Offset(slot_on_environment_stack(self, EnvironmentStackSlot::OptRetValPtr))).emit(self)?;
-        X86Instruction::store_immediate(OperandSize::S64, R10, X86IndirectAccess::Offset(0), 1).emit(self)?; // is_err = true;
-        emit_alu(self, OperandSize::S64, 0x81, 0, R11, ebpf::ELF_INSN_DUMP_OFFSET as i64 - 1, None)?;
-        X86Instruction::store(OperandSize::S64, R11, R10, X86IndirectAccess::Offset(16)).emit(self)?; // pc = self.pc + ebpf::ELF_INSN_DUMP_OFFSET;
+        // Validate that we did not reach the instruction meter limit before the exception occured
+        if self.config.enable_instruction_meter {
+            emit_validate_instruction_count(self, false, None)?;
+        }
+        emit_profile_instruction_count_of_exception(self, true)?;
         emit_jmp(self, TARGET_PC_EPILOGUE)?;
 
         // Handler for syscall exceptions
-        set_anchor(self, TARGET_PC_SYSCALL_EXCEPTION);
-        emit_profile_instruction_count_of_exception(self)?;
+        set_anchor(self, TARGET_PC_RUST_EXCEPTION);
+        emit_profile_instruction_count_of_exception(self, false)?;
         emit_jmp(self, TARGET_PC_EPILOGUE)
     }
 
@@ -1413,6 +1550,11 @@ impl JitCompiler {
         // Save instruction meter
         X86Instruction::push(ARGUMENT_REGISTERS[3]).emit(self)?;
 
+        // Initialize stop watch
+        emit_alu(self, OperandSize::S64, 0x31, R11, R11, 0, None)?; // R11 ^= R11; 
+        X86Instruction::push(R11).emit(self)?;
+        X86Instruction::push(R11).emit(self)?;
+
         // Initialize frame pointer
         X86Instruction::mov(OperandSize::S64, RSP, RBP).emit(self)?;
         emit_alu(self, OperandSize::S64, 0x81, 0, RBP, 8 * (EnvironmentStackSlot::SlotCount as i64 - 1 + self.environment_stack_key as i64), None)?;
@@ -1441,6 +1583,17 @@ impl JitCompiler {
         // Epilogue
         set_anchor(self, TARGET_PC_EPILOGUE);
 
+        // Print stop watch value
+        fn stopwatch_result(numerator: u64, denominator: u64) {
+            println!("Stop watch: {} / {} = {}", numerator, denominator, if denominator == 0 { 0.0 } else { numerator as f64 / denominator as f64 });
+        }
+        if self.stopwatch_is_active {
+            emit_rust_call(self, stopwatch_result as *const u8, &[
+                Argument { index: 1, value: Value::RegisterIndirect(RBP, slot_on_environment_stack(self, EnvironmentStackSlot::StopwatchDenominator), false) },
+                Argument { index: 0, value: Value::RegisterIndirect(RBP, slot_on_environment_stack(self, EnvironmentStackSlot::StopwatchNumerator), false) },
+            ], None, false)?;
+        }
+
         // Store instruction_meter in RAX
         X86Instruction::mov(OperandSize::S64, ARGUMENT_REGISTERS[0], RAX).emit(self)?;
 
@@ -1456,7 +1609,7 @@ impl JitCompiler {
     }
 
     pub fn emit_random_noop<E: UserDefinedError>(&mut self) -> Result<(), EbpfError<E>> {
-        if self.config.noop_instruction_ratio != 0.0 && self.rng.gen_bool(self.config.noop_instruction_ratio) {
+        if self.config.noop_instruction_ratio != 0.0 && self.diversification_rng.gen_bool(self.config.noop_instruction_ratio) {
             // X86Instruction::noop().emit(self)
             emit::<u8, E>(self, 0x90)
         } else {
