@@ -21,7 +21,7 @@
 //!
 //! Contrary to the verifier of the Linux kernel, this one does not modify the bytecode at all.
 
-use crate::ebpf;
+use crate::{ebpf, vm::Config};
 use thiserror::Error;
 
 /// Error definitions
@@ -70,8 +70,8 @@ pub enum VerifierError {
     #[error("unknown eBPF opcode {0:#2x} (insn #{1:?})")]
     UnknownOpCode(u8, usize),
     /// Shift with overflow
-    #[error("Shift with overflow at instruction {0}")]
-    ShiftWithOverflow(usize),
+    #[error("Shift with overflow of {0}-bit value by {1} (insn #{2:?})")]
+    ShiftWithOverflow(u64, u64, usize),
     /// Invalid register specified
     #[error("Invalid register specified at instruction {0}")]
     InvalidRegister(usize),
@@ -121,9 +121,6 @@ fn check_load_dw(prog: &[u8], insn_ptr: usize) -> Result<(), VerifierError> {
 
 fn check_jmp_offset(prog: &[u8], insn_ptr: usize) -> Result<(), VerifierError> {
     let insn = ebpf::get_insn(prog, insn_ptr);
-    // if insn.off == -1 {
-    //     return Err(VerifierError::InfiniteLoop(adj_insn_ptr(insn_ptr)).into());
-    // }
 
     let dst_insn_ptr = insn_ptr as isize + 1 + insn.off as isize;
     if dst_insn_ptr < 0 || dst_insn_ptr as usize >= (prog.len() / ebpf::INSN_SIZE) {
@@ -156,9 +153,14 @@ fn check_registers(insn: &ebpf::Insn, store: bool, insn_ptr: usize) -> Result<()
 }
 
 /// Check that the imm is a valid shift operand
-fn check_imm_shift(insn: &ebpf::Insn, insn_ptr: usize) -> Result<(), VerifierError> {
-    if insn.imm < 0 || insn.imm as u64 >= 64 {
-        return Err(VerifierError::ShiftWithOverflow(adj_insn_ptr(insn_ptr)));
+fn check_imm_shift(insn: &ebpf::Insn, insn_ptr: usize, imm_bits: u64) -> Result<(), VerifierError> {
+    let shift_by = insn.imm as u64;
+    if insn.imm < 0 || shift_by >= imm_bits {
+        return Err(VerifierError::ShiftWithOverflow(
+            shift_by,
+            imm_bits,
+            adj_insn_ptr(insn_ptr),
+        ));
     }
     Ok(())
 }
@@ -173,7 +175,7 @@ fn check_imm_register(insn: &ebpf::Insn, insn_ptr: usize) -> Result<(), Verifier
 
 /// Check the program against the verifier's rules
 #[rustfmt::skip]
-pub fn check(prog: &[u8]) -> Result<(), VerifierError> {
+pub fn check(prog: &[u8], config: &Config) -> Result<(), VerifierError> {
     check_prog_len(prog)?;
 
     let mut insn_ptr: usize = 0;
@@ -229,9 +231,9 @@ pub fn check(prog: &[u8]) -> Result<(), VerifierError> {
             ebpf::OR32_REG   => {},
             ebpf::AND32_IMM  => {},
             ebpf::AND32_REG  => {},
-            ebpf::LSH32_IMM  => { check_imm_shift(&insn, insn_ptr)?; },
+            ebpf::LSH32_IMM  => { check_imm_shift(&insn, insn_ptr, if config.verify_shift32_imm { 32 } else { 64 })?; },
             ebpf::LSH32_REG  => {},
-            ebpf::RSH32_IMM  => { check_imm_shift(&insn, insn_ptr)?; },
+            ebpf::RSH32_IMM  => { check_imm_shift(&insn, insn_ptr, if config.verify_shift32_imm { 32 } else { 64 })?; },
             ebpf::RSH32_REG  => {},
             ebpf::NEG32      => {},
             ebpf::MOD32_IMM  => { check_imm_nonzero(&insn, insn_ptr)?; },
@@ -240,7 +242,7 @@ pub fn check(prog: &[u8]) -> Result<(), VerifierError> {
             ebpf::XOR32_REG  => {},
             ebpf::MOV32_IMM  => {},
             ebpf::MOV32_REG  => {},
-            ebpf::ARSH32_IMM => { check_imm_shift(&insn, insn_ptr)?; },
+            ebpf::ARSH32_IMM => { check_imm_shift(&insn, insn_ptr, if config.verify_shift32_imm { 32 } else { 64 })?; },
             ebpf::ARSH32_REG => {},
             ebpf::LE         => { check_imm_endian(&insn, insn_ptr)?; },
             ebpf::BE         => { check_imm_endian(&insn, insn_ptr)?; },
@@ -250,7 +252,11 @@ pub fn check(prog: &[u8]) -> Result<(), VerifierError> {
             ebpf::ADD64_REG  => {},
             ebpf::SUB64_IMM  => {},
             ebpf::SUB64_REG  => {},
-            ebpf::MUL64_IMM  => { check_imm_nonzero(&insn, insn_ptr)?; },
+            ebpf::MUL64_IMM  => {
+                if config.verify_mul64_imm_nonzero {
+                    check_imm_nonzero(&insn, insn_ptr)?;
+                }
+            },
             ebpf::MUL64_REG  => {},
             ebpf::DIV64_IMM  => { check_imm_nonzero(&insn, insn_ptr)?; },
             ebpf::DIV64_REG  => {},
@@ -258,9 +264,9 @@ pub fn check(prog: &[u8]) -> Result<(), VerifierError> {
             ebpf::OR64_REG   => {},
             ebpf::AND64_IMM  => {},
             ebpf::AND64_REG  => {},
-            ebpf::LSH64_IMM  => { check_imm_shift(&insn, insn_ptr)?; },
+            ebpf::LSH64_IMM  => { check_imm_shift(&insn, insn_ptr, 64)?; },
             ebpf::LSH64_REG  => {},
-            ebpf::RSH64_IMM  => { check_imm_shift(&insn, insn_ptr)?; },
+            ebpf::RSH64_IMM  => { check_imm_shift(&insn, insn_ptr, 64)?; },
             ebpf::RSH64_REG  => {},
             ebpf::NEG64      => {},
             ebpf::MOD64_IMM  => { check_imm_nonzero(&insn, insn_ptr)?; },
@@ -269,7 +275,7 @@ pub fn check(prog: &[u8]) -> Result<(), VerifierError> {
             ebpf::XOR64_REG  => {},
             ebpf::MOV64_IMM  => {},
             ebpf::MOV64_REG  => {},
-            ebpf::ARSH64_IMM => { check_imm_shift(&insn, insn_ptr)?; },
+            ebpf::ARSH64_IMM => { check_imm_shift(&insn, insn_ptr, 64)?; },
             ebpf::ARSH64_REG => {},
 
             // BPF_JMP class
