@@ -37,7 +37,7 @@ use std::{
 ///   - Unknown instructions.
 ///   - Bad formed instruction.
 ///   - Unknown eBPF syscall index.
-pub type Verifier = fn(prog: &[u8]) -> Result<(), VerifierError>;
+pub type Verifier = fn(prog: &[u8], config: &Config) -> Result<(), VerifierError>;
 
 /// Return value of programs and syscalls
 pub type ProgramResult<E> = Result<u64, EbpfError<E>>;
@@ -179,6 +179,10 @@ pub struct Config {
     pub max_call_depth: usize,
     /// Size of a stack frame in bytes, must match the size specified in the LLVM BPF backend
     pub stack_frame_size: usize,
+    /// Enables gaps in VM address space between the stack frames
+    pub enable_stack_frame_gaps: bool,
+    /// Maximal pc distance after which a new instruction meter validation is emitted by the JIT
+    pub instruction_meter_checkpoint_distance: usize,
     /// Enable instruction meter and limiting
     pub enable_instruction_meter: bool,
     /// Enable instruction tracing
@@ -191,18 +195,26 @@ pub struct Config {
     pub sanitize_user_provided_values: bool,
     /// Encrypt the environment registers in JIT
     pub encrypt_environment_registers: bool,
+    /// Feature flag for the MUL64_IMM != 0 verification check
+    pub verify_mul64_imm_nonzero: bool,
+    /// Feature flag for the SHIFT_IMM >= 32 verification check
+    pub verify_shift32_imm: bool,
 }
 impl Default for Config {
     fn default() -> Self {
         Self {
             max_call_depth: 20,
             stack_frame_size: 4_096,
+            enable_stack_frame_gaps: true,
+            instruction_meter_checkpoint_distance: 10000,
             enable_instruction_meter: true,
             enable_instruction_tracing: false,
             reject_unresolved_syscalls: false,
             noop_instruction_ratio: 1.0 / 256.0,
             sanitize_user_provided_values: true,
             encrypt_environment_registers: true,
+            verify_mul64_imm_nonzero: false,
+            verify_shift32_imm: false,
         }
     }
 }
@@ -212,9 +224,9 @@ pub trait Executable<E: UserDefinedError, I: InstructionMeter>: Send + Sync {
     /// Get the configuration settings
     fn get_config(&self) -> &Config;
     /// Get the .text section virtual address and bytes
-    fn get_text_bytes(&self) -> Result<(u64, &[u8]), EbpfError<E>>;
-    /// Get a vector of virtual addresses for each read-only section
-    fn get_ro_sections(&self) -> Result<Vec<(u64, &[u8])>, EbpfError<E>>;
+    fn get_text_bytes(&self) -> (u64, &[u8]);
+    /// Get the concatenated read-only sections (including the text section)
+    fn get_ro_section(&self) -> &[u8];
     /// Get the entry point offset into the text section
     fn get_entrypoint_instruction_offset(&self) -> Result<usize, EbpfError<E>>;
     /// Get a symbol's instruction offset
@@ -233,11 +245,17 @@ pub trait Executable<E: UserDefinedError, I: InstructionMeter>: Send + Sync {
     fn get_syscall_symbols(&self) -> &BTreeMap<u32, String>;
 }
 
+#[cfg(vtable_send_sync_plus_one)]
+/// Index of report_unresolved_symbol in the Executable traits vtable
+///
+/// It is shifted by 1 because Send + Sync needs an extra slot.
+pub const REPORT_UNRESOLVED_SYMBOL_INDEX: usize = 9;
+#[cfg(not(vtable_send_sync_plus_one))]
 /// Index of report_unresolved_symbol in the Executable traits vtable
 pub const REPORT_UNRESOLVED_SYMBOL_INDEX: usize = 8;
 
 /// The syscall_context_objects field stores some metadata in the front, thus the entries are shifted
-pub const SYSCALL_CONTEXT_OBJECTS_OFFSET: usize = 6;
+pub const SYSCALL_CONTEXT_OBJECTS_OFFSET: usize = 4;
 
 /// Static constructors for Executable
 impl<E: UserDefinedError, I: 'static + InstructionMeter> dyn Executable<E, I> {
@@ -249,9 +267,9 @@ impl<E: UserDefinedError, I: 'static + InstructionMeter> dyn Executable<E, I> {
         syscall_registry: SyscallRegistry,
     ) -> Result<Box<Self>, EbpfError<E>> {
         let ebpf_elf = EBpfElf::load(config, elf_bytes, syscall_registry)?;
-        let (_, bytes) = ebpf_elf.get_text_bytes()?;
+        let text_bytes = ebpf_elf.get_text_bytes().1;
         if let Some(verifier) = verifier {
-            verifier(bytes)?;
+            verifier(text_bytes, &config)?;
         }
         Ok(Box::new(ebpf_elf))
     }
@@ -264,7 +282,7 @@ impl<E: UserDefinedError, I: 'static + InstructionMeter> dyn Executable<E, I> {
         bpf_functions: BTreeMap<u32, (usize, String)>,
     ) -> Result<Box<Self>, EbpfError<E>> {
         if let Some(verifier) = verifier {
-            verifier(text_bytes).map_err(EbpfError::VerifierError)?;
+            verifier(text_bytes, &config).map_err(EbpfError::VerifierError)?;
         }
         Ok(Box::new(EBpfElf::new_from_text_bytes(
             config,
@@ -283,13 +301,21 @@ pub trait InstructionMeter {
     fn get_remaining(&self) -> u64;
 }
 
-/// Instruction meter without a limit
-#[derive(Debug, PartialEq)]
-pub struct DefaultInstructionMeter {}
-impl InstructionMeter for DefaultInstructionMeter {
-    fn consume(&mut self, _amount: u64) {}
+/// Simple instruction meter for testing
+#[derive(Debug, PartialEq, Eq)]
+pub struct TestInstructionMeter {
+    /// Maximal amount of instructions which still can be executed
+    pub remaining: u64,
+}
+
+impl InstructionMeter for TestInstructionMeter {
+    fn consume(&mut self, amount: u64) {
+        debug_assert!(amount <= self.remaining, "Execution count exceeded");
+        self.remaining = self.remaining.saturating_sub(amount);
+    }
+
     fn get_remaining(&self) -> u64 {
-        std::i64::MAX as u64
+        self.remaining
     }
 }
 
@@ -427,7 +453,7 @@ macro_rules! translate_memory_access {
 /// # Examples
 ///
 /// ```
-/// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, DefaultInstructionMeter, SyscallRegistry}, user_error::UserError};
+/// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, TestInstructionMeter, SyscallRegistry}, user_error::UserError};
 ///
 /// let prog = &[
 ///     0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // exit
@@ -439,11 +465,11 @@ macro_rules! translate_memory_access {
 /// // Instantiate a VM.
 /// let mut bpf_functions = std::collections::BTreeMap::new();
 /// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
-/// let mut executable = <dyn Executable::<UserError, DefaultInstructionMeter>>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
-/// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), mem, &[]).unwrap();
+/// let mut executable = <dyn Executable::<UserError, TestInstructionMeter>>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
+/// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(executable.as_ref(), &mut [], mem).unwrap();
 ///
 /// // Provide a reference to the packet data.
-/// let res = vm.execute_program_interpreted(&mut DefaultInstructionMeter {}).unwrap();
+/// let res = vm.execute_program_interpreted(&mut TestInstructionMeter { remaining: 1 }).unwrap();
 /// assert_eq!(res, 0);
 /// ```
 pub struct EbpfVm<'a, E: UserDefinedError, I: InstructionMeter> {
@@ -454,7 +480,7 @@ pub struct EbpfVm<'a, E: UserDefinedError, I: InstructionMeter> {
     tracer: Tracer,
     syscall_context_objects: Vec<*mut u8>,
     syscall_context_object_pool: Vec<Box<dyn SyscallObject<E> + 'a>>,
-    frames: CallFrames,
+    stack: CallFrames<'a>,
     last_insn_count: u64,
     total_insn_count: u64,
 }
@@ -466,7 +492,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// # Examples
     ///
     /// ```
-    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, DefaultInstructionMeter, SyscallRegistry}, user_error::UserError};
+    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, TestInstructionMeter, SyscallRegistry}, user_error::UserError};
     ///
     /// let prog = &[
     ///     0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // exit
@@ -475,43 +501,25 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// // Instantiate a VM.
     /// let mut bpf_functions = std::collections::BTreeMap::new();
     /// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
-    /// let mut executable = <dyn Executable::<UserError, DefaultInstructionMeter>>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
-    /// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), &mut [], &[]).unwrap();
+    /// let mut executable = <dyn Executable::<UserError, TestInstructionMeter>>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
+    /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(executable.as_ref(), &mut [], &mut []).unwrap();
     /// ```
     pub fn new(
         executable: &'a dyn Executable<E, I>,
-        mem: &mut [u8],
-        granted_regions: &[MemoryRegion],
+        heap_region: &mut [u8],
+        input_region: &mut [u8],
     ) -> Result<EbpfVm<'a, E, I>, EbpfError<E>> {
         let config = executable.get_config();
-        let const_data_regions: Vec<MemoryRegion> =
-            if let Ok(sections) = executable.get_ro_sections() {
-                sections
-                    .iter()
-                    .map(|(addr, slice)| MemoryRegion::new_from_slice(slice, *addr, 0, false))
-                    .collect()
-            } else {
-                Vec::new()
-            };
-        let mut regions: Vec<MemoryRegion> =
-            Vec::with_capacity(granted_regions.len() + const_data_regions.len() + 3);
-        regions.extend(granted_regions.iter().cloned());
-        let frames = CallFrames::new(config.max_call_depth, config.stack_frame_size);
-        regions.push(frames.get_region().clone());
-        regions.extend(const_data_regions);
-        regions.push(MemoryRegion::new_from_slice(
-            mem,
-            ebpf::MM_INPUT_START,
-            0,
-            true,
-        ));
-        let (program_vm_addr, program) = executable.get_text_bytes()?;
-        regions.push(MemoryRegion::new_from_slice(
-            program,
-            program_vm_addr,
-            0,
-            false,
-        ));
+        let ro_region = executable.get_ro_section();
+        let stack = CallFrames::new(config);
+        let regions: Vec<MemoryRegion> = vec![
+            MemoryRegion::new_from_slice(&[], 0, 0, false),
+            MemoryRegion::new_from_slice(ro_region, ebpf::MM_PROGRAM_START, 0, false),
+            stack.get_memory_region(),
+            MemoryRegion::new_from_slice(heap_region, ebpf::MM_HEAP_START, 0, true),
+            MemoryRegion::new_from_slice(input_region, ebpf::MM_INPUT_START, 0, true),
+        ];
+        let (program_vm_addr, program) = executable.get_text_bytes();
         let number_of_syscalls = executable.get_syscall_registry().get_number_of_syscalls();
         let mut vm = EbpfVm {
             executable,
@@ -524,7 +532,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 SYSCALL_CONTEXT_OBJECTS_OFFSET + number_of_syscalls
             ],
             syscall_context_object_pool: Vec::with_capacity(number_of_syscalls),
-            frames,
+            stack,
             last_insn_count: 0,
             total_insn_count: 0,
         };
@@ -558,7 +566,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// # Examples
     ///
     /// ```
-    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, SyscallObject, SyscallRegistry, DefaultInstructionMeter}, syscalls::BpfTracePrintf, user_error::UserError};
+    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, SyscallObject, SyscallRegistry, TestInstructionMeter}, syscalls::BpfTracePrintf, user_error::UserError};
     ///
     /// // This program was compiled with clang, from a C program containing the following single
     /// // instruction: `return bpf_trace_printk("foo %c %c %c\n", 10, 1, 2, 3);`
@@ -583,8 +591,8 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// // Instantiate an Executable and VM
     /// let mut bpf_functions = std::collections::BTreeMap::new();
     /// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
-    /// let mut executable = <dyn Executable::<UserError, DefaultInstructionMeter>>::from_text_bytes(prog, None, Config::default(), syscall_registry, bpf_functions).unwrap();
-    /// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), &mut [], &[]).unwrap();
+    /// let mut executable = <dyn Executable::<UserError, TestInstructionMeter>>::from_text_bytes(prog, None, Config::default(), syscall_registry, bpf_functions).unwrap();
+    /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(executable.as_ref(), &mut [], &mut []).unwrap();
     /// // Bind a context object instance to the previously registered syscall
     /// vm.bind_syscall_context_object(Box::new(BpfTracePrintf {}), None);
     /// ```
@@ -635,7 +643,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// # Examples
     ///
     /// ```
-    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, DefaultInstructionMeter, SyscallRegistry}, user_error::UserError};
+    /// use solana_rbpf::{ebpf, elf::register_bpf_function, vm::{Config, Executable, EbpfVm, TestInstructionMeter, SyscallRegistry}, user_error::UserError};
     ///
     /// let prog = &[
     ///     0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00  // exit
@@ -647,11 +655,11 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
     /// // Instantiate a VM.
     /// let mut bpf_functions = std::collections::BTreeMap::new();
     /// register_bpf_function(&mut bpf_functions, 0, "entrypoint").unwrap();
-    /// let mut executable = <dyn Executable::<UserError, DefaultInstructionMeter>>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
-    /// let mut vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), mem, &[]).unwrap();
+    /// let mut executable = <dyn Executable::<UserError, TestInstructionMeter>>::from_text_bytes(prog, None, Config::default(), SyscallRegistry::default(), bpf_functions).unwrap();
+    /// let mut vm = EbpfVm::<UserError, TestInstructionMeter>::new(executable.as_ref(), &mut [], mem).unwrap();
     ///
     /// // Provide a reference to the packet data.
-    /// let res = vm.execute_program_interpreted(&mut DefaultInstructionMeter {}).unwrap();
+    /// let res = vm.execute_program_interpreted(&mut TestInstructionMeter { remaining: 1 }).unwrap();
     /// assert_eq!(res, 0);
     /// ```
     pub fn execute_program_interpreted(&mut self, instruction_meter: &mut I) -> ProgramResult<E> {
@@ -676,7 +684,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
         const U32MAX: u64 = u32::MAX as u64;
 
         // R1 points to beginning of input memory, R10 to the stack of the first frame
-        let mut reg: [u64; 11] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, self.frames.get_stack_top()];
+        let mut reg: [u64; 11] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, self.stack.get_stack_top()];
 
         if self.memory_mapping.map::<UserError>(AccessType::Store, ebpf::MM_INPUT_START, 1).is_ok() {
             reg[1] = ebpf::MM_INPUT_START;
@@ -948,7 +956,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 ebpf::CALL_REG   => {
                     let target_address = reg[insn.imm as usize];
                     reg[ebpf::STACK_REG] =
-                        self.frames.push(&reg[ebpf::FIRST_SCRATCH_REG..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS], next_pc)?;
+                        self.stack.push(&reg[ebpf::FIRST_SCRATCH_REG..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS], next_pc)?;
                     if target_address < self.program_vm_addr {
                         return Err(EbpfError::CallOutsideTextSegment(pc + ebpf::ELF_INSN_DUMP_OFFSET, target_address / ebpf::INSN_SIZE as u64 * ebpf::INSN_SIZE as u64));
                     }
@@ -981,7 +989,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                         }
                     } else if let Some(target_pc) = self.executable.lookup_bpf_function(insn.imm as u32) {
                         // make BPF to BPF call
-                        reg[ebpf::STACK_REG] = self.frames.push(
+                        reg[ebpf::STACK_REG] = self.stack.push(
                             &reg[ebpf::FIRST_SCRATCH_REG
                                 ..ebpf::FIRST_SCRATCH_REG + ebpf::SCRATCH_REGS],
                             next_pc,
@@ -993,7 +1001,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                 }
 
                 ebpf::EXIT => {
-                    match self.frames.pop::<E>() {
+                    match self.stack.pop::<E>() {
                         Ok((saved_reg, stack_ptr, ptr)) => {
                             // Return from BPF to BPF call
                             reg[ebpf::FIRST_SCRATCH_REG
@@ -1006,7 +1014,7 @@ impl<'a, E: UserDefinedError, I: InstructionMeter> EbpfVm<'a, E, I> {
                             debug!("BPF instructions executed (interp): {:?}", total_insn_count + self.last_insn_count);
                             debug!(
                                 "Max frame depth reached: {:?}",
-                                self.frames.get_max_frame_index()
+                                self.stack.get_max_frame_index()
                             );
                             return Ok(reg[0]);
                         }
