@@ -10,21 +10,20 @@ extern crate solana_rbpf;
 extern crate test_utils;
 extern crate thiserror;
 
-#[cfg(not(windows))]
+#[cfg(all(not(windows), target_arch = "x86_64"))]
 use rand::{rngs::SmallRng, RngCore, SeedableRng};
 use solana_rbpf::{
     assembler::assemble,
     elf::ElfError,
     error::EbpfError,
     memory_region::AccessType,
+    memory_region::MemoryMapping,
     syscalls::{self, Result},
     user_error::UserError,
-    vm::{Config, EbpfVm, Executable, SyscallObject, SyscallRegistry},
+    vm::{Config, EbpfVm, Executable, SyscallObject, SyscallRegistry, TestInstructionMeter},
 };
 use std::{fs::File, io::Read};
-use test_utils::{
-    TestInstructionMeter, PROG_TCP_PORT_80, TCP_SACK_ASM, TCP_SACK_MATCH, TCP_SACK_NOMATCH,
-};
+use test_utils::{PROG_TCP_PORT_80, TCP_SACK_ASM, TCP_SACK_MATCH, TCP_SACK_NOMATCH};
 
 macro_rules! test_interpreter_and_jit {
     (register, $syscall_registry:expr, $location:expr => $syscall_function:expr; $syscall_context_object:expr) => {
@@ -37,18 +36,18 @@ macro_rules! test_interpreter_and_jit {
         let check_closure = $check;
         let (instruction_count_interpreter, _tracer_interpreter) = {
             let mut mem = $mem;
-            let mut vm = EbpfVm::new($executable.as_ref(), &mut mem, &[]).unwrap();
+            let mut vm = EbpfVm::new($executable.as_ref(), &mut [], &mut mem).unwrap();
             $(test_interpreter_and_jit!(bind, vm, $location => $syscall_function; $syscall_context_object);)*
             let result = vm.execute_program_interpreted(&mut TestInstructionMeter { remaining: $expected_instruction_count });
             assert!(check_closure(&vm, result));
             (vm.get_total_instruction_count(), vm.get_tracer().clone())
         };
-        #[cfg(not(windows))]
+        #[cfg(all(not(windows), target_arch = "x86_64"))]
         {
             let check_closure = $check;
             let compilation_result = $executable.jit_compile();
             let mut mem = $mem;
-            let mut vm = EbpfVm::new($executable.as_ref(), &mut mem, &[]).unwrap();
+            let mut vm = EbpfVm::new($executable.as_ref(), &mut [], &mut mem).unwrap();
             match compilation_result {
                 Err(err) => assert!(check_closure(&vm, Err(err))),
                 Ok(()) => {
@@ -2703,6 +2702,84 @@ fn test_syscall_with_context() {
     );
 }
 
+pub struct NestedVmSyscall {}
+impl SyscallObject<UserError> for NestedVmSyscall {
+    fn call(
+        &mut self,
+        depth: u64,
+        throw: u64,
+        _arg3: u64,
+        _arg4: u64,
+        _arg5: u64,
+        _memory_mapping: &MemoryMapping,
+        result: &mut Result,
+    ) {
+        #[allow(unused_mut)]
+        if depth > 0 {
+            let mut syscall_registry = SyscallRegistry::default();
+            syscall_registry
+                .register_syscall_by_name::<UserError, _>(b"NestedVmSyscall", NestedVmSyscall::call)
+                .unwrap();
+            let mem = &mut [depth as u8 - 1, throw as u8];
+            let mut executable = assemble::<UserError, TestInstructionMeter>(
+                "
+                ldabsb 0
+                mov64 r1, r0
+                ldabsb 1
+                mov64 r2, r0
+                syscall NestedVmSyscall
+                exit",
+                None,
+                Config::default(),
+                syscall_registry,
+            )
+            .unwrap();
+            #[cfg(all(not(windows), target_arch = "x86_64"))]
+            {
+                executable.jit_compile().unwrap();
+            }
+            let mut vm = EbpfVm::new(executable.as_ref(), &mut [], mem).unwrap();
+            vm.bind_syscall_context_object(Box::new(NestedVmSyscall {}), None)
+                .unwrap();
+            let mut instruction_meter = TestInstructionMeter { remaining: 6 };
+            #[cfg(any(windows, not(target_arch = "x86_64")))]
+            {
+                *result = vm.execute_program_interpreted(&mut instruction_meter);
+            }
+            #[cfg(all(not(windows), target_arch = "x86_64"))]
+            {
+                *result = vm.execute_program_jit(&mut instruction_meter);
+            }
+            assert_eq!(
+                vm.get_total_instruction_count(),
+                if throw == 0 { 6 } else { 5 }
+            );
+        } else {
+            *result = if throw == 0 {
+                Ok(42)
+            } else {
+                Err(EbpfError::CallDepthExceeded(33, 0))
+            };
+        }
+    }
+}
+
+#[test]
+fn test_nested_vm_syscall() {
+    let config = Config::default();
+    let mut nested_vm_syscall = NestedVmSyscall {};
+    let memory_mapping = MemoryMapping::new::<UserError>(vec![], &config).unwrap();
+    let mut result = Ok(0);
+    nested_vm_syscall.call(1, 0, 0, 0, 0, &memory_mapping, &mut result);
+    assert!(result.unwrap() == 42);
+    let mut result = Ok(0);
+    nested_vm_syscall.call(1, 1, 0, 0, 0, &memory_mapping, &mut result);
+    assert!(matches!(result.unwrap_err(),
+        EbpfError::CallDepthExceeded(pc, depth)
+        if pc == 33 && depth == 0
+    ));
+}
+
 // Elf
 
 #[test]
@@ -2995,6 +3072,53 @@ fn test_err_non_terminating_capped() {
     );
 }
 
+#[test]
+fn test_err_capped_before_exception() {
+    test_interpreter_and_jit_asm!(
+        "
+        mov64 r1, 0x0
+        mov64 r2, 0x0
+        add64 r0, 0x0
+        add64 r0, 0x0
+        div64 r1, r2
+        add64 r0, 0x0
+        exit",
+        [],
+        (),
+        {
+            |_vm, res: Result| {
+                matches!(res.unwrap_err(),
+                    EbpfError::ExceededMaxInstructions(pc, initial_insn_count)
+                    if pc == 31 && initial_insn_count == 2
+                )
+            }
+        },
+        2
+    );
+
+    test_interpreter_and_jit_asm!(
+        "
+        mov64 r1, 0x0
+        mov64 r2, 0x0
+        add64 r0, 0x0
+        add64 r0, 0x0
+        syscall Unresolved
+        add64 r0, 0x0
+        exit",
+        [],
+        (),
+        {
+            |_vm, res: Result| {
+                matches!(res.unwrap_err(),
+                    EbpfError::ExceededMaxInstructions(pc, initial_insn_count)
+                    if pc == 33 && initial_insn_count == 4
+                )
+            }
+        },
+        4
+    );
+}
+
 // Symbols and Relocation
 
 #[test]
@@ -3016,22 +3140,6 @@ fn test_symbol_relocation() {
     );
 }
 
-#[test]
-fn test_err_symbol_unresolved() {
-    test_interpreter_and_jit_asm!(
-        "
-        syscall Unresolved
-        mov64 r0, 0x0
-        exit",
-        [],
-        (),
-        {
-            |_vm, res: Result| matches!(res.unwrap_err(), EbpfError::ElfError(ElfError::UnresolvedSymbol(symbol, pc, offset)) if symbol == "Unknown" && pc == 29 && offset == 0)
-        },
-        1
-    );
-}
-
 #[test]
 fn test_err_call_unresolved() {
     test_interpreter_and_jit_asm!(
@@ -3042,6 +3150,7 @@ fn test_err_call_unresolved() {
         mov r4, 4
         mov r5, 5
         syscall Unresolved
+        mov64 r0, 0x0
         exit",
         [],
         (),
@@ -3061,7 +3170,7 @@ fn test_err_unresolved_elf() {
     file.read_to_end(&mut elf).unwrap();
     let config = Config {
         reject_unresolved_syscalls: true,
-        ..Default::default()
+        ..Config::default()
     };
     assert!(
         matches!(<dyn Executable::<UserError, TestInstructionMeter>>::from_elf(&elf, None, config, syscall_registry), Err(EbpfError::ElfError(ElfError::UnresolvedSymbol(symbol, pc, offset))) if symbol == "log_64" && pc == 550 && offset == 4168)
@@ -3278,7 +3387,7 @@ fn test_tcp_sack_nomatch() {
 
 // Fuzzy
 
-#[cfg(not(windows))]
+#[cfg(all(not(windows), target_arch = "x86_64"))]
 fn execute_generated_program(prog: &[u8]) -> bool {
     use solana_rbpf::{elf::register_bpf_function, verifier::check};
     use std::collections::BTreeMap;
@@ -3307,7 +3416,7 @@ fn execute_generated_program(prog: &[u8]) -> bool {
     }
     let (instruction_count_interpreter, tracer_interpreter, result_interpreter) = {
         let mut mem = vec![0u8; mem_size];
-        let mut vm = EbpfVm::new(executable.as_ref(), &mut mem, &[]).unwrap();
+        let mut vm = EbpfVm::new(executable.as_ref(), &mut [], &mut mem).unwrap();
         let result_interpreter = vm.execute_program_interpreted(&mut TestInstructionMeter {
             remaining: max_instruction_count,
         });
@@ -3319,7 +3428,7 @@ fn execute_generated_program(prog: &[u8]) -> bool {
         )
     };
     let mut mem = vec![0u8; mem_size];
-    let mut vm = EbpfVm::new(executable.as_ref(), &mut mem, &[]).unwrap();
+    let mut vm = EbpfVm::new(executable.as_ref(), &mut [], &mut mem).unwrap();
     let result_jit = vm.execute_program_jit(&mut TestInstructionMeter {
         remaining: max_instruction_count,
     });
@@ -3344,7 +3453,7 @@ fn execute_generated_program(prog: &[u8]) -> bool {
     true
 }
 
-#[cfg(not(windows))]
+#[cfg(all(not(windows), target_arch = "x86_64"))]
 #[test]
 fn test_total_chaos() {
     use solana_rbpf::ebpf;
