@@ -27,7 +27,7 @@ use solana_rbpf::{
     error::UserDefinedError,
     user_error::UserError,
     verifier::{check, VerifierError},
-    vm::{Config, DefaultInstructionMeter, EbpfVm, Executable, SyscallRegistry},
+    vm::{Config, EbpfVm, Executable, SyscallRegistry, TestInstructionMeter},
 };
 use std::collections::BTreeMap;
 use thiserror::Error;
@@ -42,26 +42,26 @@ impl UserDefinedError for VerifierTestError {}
 
 #[test]
 fn test_verifier_success() {
-    let executable = assemble::<UserError, DefaultInstructionMeter>(
+    let executable = assemble::<UserError, TestInstructionMeter>(
         "
         mov32 r0, 0xBEE
         exit",
-        Some(|_prog: &[u8]| Ok(())),
+        Some(|_prog: &[u8], _config: &Config| Ok(())),
         Config::default(),
         SyscallRegistry::default(),
     )
     .unwrap();
-    let _vm = EbpfVm::<UserError, DefaultInstructionMeter>::new(executable.as_ref(), &mut [], &[])
+    let _vm = EbpfVm::<UserError, TestInstructionMeter>::new(executable.as_ref(), &mut [], &mut [])
         .unwrap();
 }
 
 #[test]
 #[should_panic(expected = "NoProgram")]
 fn test_verifier_fail() {
-    fn verifier_fail(_prog: &[u8]) -> Result<(), VerifierError> {
+    fn verifier_fail(_prog: &[u8], _config: &Config) -> Result<(), VerifierError> {
         Err(VerifierError::NoProgram)
     }
-    let _executable = assemble::<UserError, DefaultInstructionMeter>(
+    let _executable = assemble::<UserError, TestInstructionMeter>(
         "
         mov32 r0, 0xBEE
         exit",
@@ -75,7 +75,7 @@ fn test_verifier_fail() {
 #[test]
 #[should_panic(expected = "DivisionByZero(30)")]
 fn test_verifier_err_div_by_zero_imm() {
-    let _executable = assemble::<UserError, DefaultInstructionMeter>(
+    let _executable = assemble::<UserError, TestInstructionMeter>(
         "
         mov32 r0, 1
         div32 r0, 0
@@ -95,7 +95,7 @@ fn test_verifier_err_endian_size() {
         0xb7, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //
         0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //
     ];
-    let _ = <dyn Executable<UserError, DefaultInstructionMeter>>::from_text_bytes(
+    let _ = <dyn Executable<UserError, TestInstructionMeter>>::from_text_bytes(
         prog,
         Some(check),
         Config::default(),
@@ -113,7 +113,7 @@ fn test_verifier_err_incomplete_lddw() {
         0x18, 0x00, 0x00, 0x00, 0x88, 0x77, 0x66, 0x55, //
         0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //
     ];
-    let _ = <dyn Executable<UserError, DefaultInstructionMeter>>::from_text_bytes(
+    let _ = <dyn Executable<UserError, TestInstructionMeter>>::from_text_bytes(
         prog,
         Some(check),
         Config::default(),
@@ -126,7 +126,7 @@ fn test_verifier_err_incomplete_lddw() {
 #[test]
 #[should_panic(expected = "InvalidDestinationRegister(29)")]
 fn test_verifier_err_invalid_reg_dst() {
-    let _executable = assemble::<UserError, DefaultInstructionMeter>(
+    let _executable = assemble::<UserError, TestInstructionMeter>(
         "
         mov r11, 1
         exit",
@@ -140,7 +140,7 @@ fn test_verifier_err_invalid_reg_dst() {
 #[test]
 #[should_panic(expected = "InvalidSourceRegister(29)")]
 fn test_verifier_err_invalid_reg_src() {
-    let _executable = assemble::<UserError, DefaultInstructionMeter>(
+    let _executable = assemble::<UserError, TestInstructionMeter>(
         "
         mov r0, r11
         exit",
@@ -154,7 +154,7 @@ fn test_verifier_err_invalid_reg_src() {
 #[test]
 #[should_panic(expected = "JumpToMiddleOfLDDW(2, 29)")]
 fn test_verifier_err_jmp_lddw() {
-    let _executable = assemble::<UserError, DefaultInstructionMeter>(
+    let _executable = assemble::<UserError, TestInstructionMeter>(
         "
         ja +1
         lddw r0, 0x1122334455667788
@@ -169,7 +169,7 @@ fn test_verifier_err_jmp_lddw() {
 #[test]
 #[should_panic(expected = "JumpOutOfCode(3, 29)")]
 fn test_verifier_err_jmp_out() {
-    let _executable = assemble::<UserError, DefaultInstructionMeter>(
+    let _executable = assemble::<UserError, TestInstructionMeter>(
         "
         ja +2
         exit",
@@ -187,7 +187,7 @@ fn test_verifier_err_unknown_opcode() {
         0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //
         0x95, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //
     ];
-    let _ = <dyn Executable<UserError, DefaultInstructionMeter>>::from_text_bytes(
+    let _ = <dyn Executable<UserError, TestInstructionMeter>>::from_text_bytes(
         prog,
         Some(check),
         Config::default(),
@@ -200,7 +200,7 @@ fn test_verifier_err_unknown_opcode() {
 #[test]
 #[should_panic(expected = "CannotWriteR10(29)")]
 fn test_verifier_err_write_r10() {
-    let _executable = assemble::<UserError, DefaultInstructionMeter>(
+    let _executable = assemble::<UserError, TestInstructionMeter>(
         "
         mov r10, 1
         exit",
@@ -210,3 +210,53 @@ fn test_verifier_err_write_r10() {
     )
     .unwrap();
 }
+
+#[test]
+fn test_verifier_err_all_shift_overflows() {
+    let testcases = [
+        // lsh32_imm
+        ("lsh32 r0, 16", Ok(())),
+        ("lsh32 r0, 32", Err("ShiftWithOverflow(32, 32, 29)")),
+        ("lsh32 r0, 64", Err("ShiftWithOverflow(64, 32, 29)")),
+        // rsh32_imm
+        ("rsh32 r0, 16", Ok(())),
+        ("rsh32 r0, 32", Err("ShiftWithOverflow(32, 32, 29)")),
+        ("rsh32 r0, 64", Err("ShiftWithOverflow(64, 32, 29)")),
+        // arsh32_imm
+        ("arsh32 r0, 16", Ok(())),
+        ("arsh32 r0, 32", Err("ShiftWithOverflow(32, 32, 29)")),
+        ("arsh32 r0, 64", Err("ShiftWithOverflow(64, 32, 29)")),
+        // lsh64_imm
+        ("lsh64 r0, 32", Ok(())),
+        ("lsh64 r0, 64", Err("ShiftWithOverflow(64, 64, 29)")),
+        // rsh64_imm
+        ("rsh64 r0, 32", Ok(())),
+        ("rsh64 r0, 64", Err("ShiftWithOverflow(64, 64, 29)")),
+        // arsh64_imm
+        ("arsh64 r0, 32", Ok(())),
+        ("arsh64 r0, 64", Err("ShiftWithOverflow(64, 64, 29)")),
+    ];
+
+    for (overflowing_instruction, expected) in testcases {
+        let assembly = format!("\n{}\nexit", overflowing_instruction);
+        let result = assemble::<UserError, TestInstructionMeter>(
+            &assembly,
+            Some(check),
+            Config {
+                verify_shift32_imm: true,
+                ..Config::default()
+            },
+            SyscallRegistry::default(),
+        );
+        match expected {
+            Ok(()) => assert!(result.is_ok()),
+            Err(overflow_msg) => match result {
+                Err(err) => assert_eq!(
+                    err,
+                    format!("Executable constructor VerifierError({})", overflow_msg),
+                ),
+                _ => panic!("Expected error"),
+            },
+        }
+    }
+}
